from dialect import _dialect

from kirin import ir
from kirin.prelude import basic_no_opt


@ir.dialect_group(basic_no_opt.add(dialect=_dialect))
def pauli(self):
    def run_pass(mt):
        # TODO
        pass

    return run_pass



from dataclasses import field, dataclass

from kirin.ir.method import Method
from kirin.passes.fold import Fold
from kirin.rewrite.abc import RewriteResult
from kirin.passes.aggressive import Fold as AggressiveFold

from .abc import Pass
from .typeinfer import TypeInfer
from .hint_const import HintConst
from .canonicalize import Canonicalize


@dataclass
class Default(Pass):
    verify: bool = field(default=True, kw_only=True)
    fold: bool = field(default=True, kw_only=True)
    aggressive: bool = field(default=False, kw_only=True)
    typeinfer: bool = field(default=True, kw_only=True)

    canonicalize: Canonicalize = field(init=False)
    hint_const_pass: HintConst = field(init=False)
    typeinfer_pass: TypeInfer = field(init=False)
    fold_pass: Pass = field(init=False)

    def __post_init__(self):
        # TODO: cleanup no_raise
        self.typeinfer_pass = TypeInfer(self.dialects, no_raise=self.no_raise)
        self.canonicalize = Canonicalize(self.dialects, no_raise=self.no_raise)

        self.hint_const_pass = HintConst(self.dialects, no_raise=self.no_raise)
        if self.aggressive:
            self.fold_pass = AggressiveFold(self.dialects, no_raise=self.no_raise)
        else:
            self.fold_pass = Fold(self.dialects, no_raise=self.no_raise)

    def unsafe_run(self, mt: Method) -> RewriteResult:
        if self.verify:
            mt.verify()

        result = self.canonicalize.fixpoint(mt)
        if self.typeinfer:
            result = self.typeinfer_pass(mt).join(result)
            if self.verify:
                mt.verify_type()

        if self.fold:
            if self.aggressive:
                result = self.fold_pass.fixpoint(mt).join(result)
            else:
                result = self.fold_pass(mt).join(result)
        return result



from dataclasses import field, dataclass

from kirin.ir import Method
from kirin.passes import Fold, Pass, TypeInfer
from kirin.rewrite import Walk
from kirin.rewrite.abc import RewriteResult
from kirin.dialects.scf.unroll import ForLoop, PickIfElse


@dataclass
class UnrollScf(Pass):
    """This pass can be used to unroll scf.For loops and inline/expand scf.IfElse when
    the input are known at compile time.

    usage:
        UnrollScf(dialects).fixpoint(method)

    Note: This pass should be used in a fixpoint manner, to unroll nested scf nodes.

    """

    typeinfer: TypeInfer = field(init=False)
    fold: Fold = field(init=False)

    def __post_init__(self):
        self.typeinfer = TypeInfer(self.dialects, no_raise=self.no_raise)
        self.fold = Fold(self.dialects, no_raise=self.no_raise)

    def unsafe_run(self, mt: Method):
        result = RewriteResult()
        result = Walk(PickIfElse()).rewrite(mt.code).join(result)
        result = Walk(ForLoop()).rewrite(mt.code).join(result)
        result = self.typeinfer(mt).join(result)
        result = self.fold(mt).join(result)
        return result



from __future__ import annotations

from typing import TYPE_CHECKING, final

from kirin import ir, types, interp
from kirin.decl import fields
from kirin.analysis.forward import Forward, ForwardFrame

from .solve import TypeResolution

if TYPE_CHECKING:
    from kirin.dialects.func.attrs import Signature


@final
class TypeInference(Forward[types.TypeAttribute]):
    """Type inference analysis for kirin.

    This analysis uses the forward dataflow analysis framework to infer the types of
    the IR. The analysis uses the type information within the IR to determine the
    method dispatch.

    The analysis will fallback to a type resolution algorithm if the type information
    is not available in the IR but the type information is available in the abstract
    values.
    """

    keys = ("typeinfer",)
    lattice = types.TypeAttribute

    def run(self, method: ir.Method, *args, **kwargs):
        if not args and not kwargs:  # no args or kwargs
            # use the method signature to get the args
            args = method.arg_types
        return super().run(method, *args, **kwargs)

    def method_self(self, method: ir.Method) -> types.TypeAttribute:
        return method.self_type

    def frame_call(
        self,
        frame: ForwardFrame[types.TypeAttribute],
        node: ir.Statement,
        *args: types.TypeAttribute,
        **kwargs: types.TypeAttribute,
    ) -> types.TypeAttribute:
        trait = node.get_present_trait(ir.CallableStmtInterface)
        args = trait.align_input_args(node, *args, **kwargs)
        region = trait.get_callable_region(node)
        if self.state.depth >= self.max_depth:
            return self.recursion_limit_reached()

        if trait := node.get_trait(ir.HasSignature):
            signature: Signature[types.TypeAttribute] | None = trait.get_signature(node)
            args = (args[0],) + tuple(
                input.meet(arg) for input, arg in zip(signature.inputs, args[1:])
            )
        else:
            signature = None
        ret = self.frame_call_region(frame, node, region, *args)

        if isinstance(ret, interp.ReturnValue):
            return ret.value if signature is None else ret.value.meet(signature.output)
        elif not ret:  # empty result or None
            return self.void if signature is None else self.void.meet(signature.output)
        raise interp.InterpreterError(
            f"callable region {node.name} does not return `ReturnValue`, got {ret}"
        )

    def eval_fallback(
        self, frame: ForwardFrame[types.TypeAttribute], node: ir.Statement
    ) -> interp.StatementResult[types.TypeAttribute]:
        resolve = TypeResolution()
        fs = fields(node)
        for f, value in zip(fs.args.values(), frame.get_values(node.args)):
            resolve.solve(f.type, value)
        for arg, f in zip(node.args, fs.args.values()):
            frame.set(arg, frame.get(arg).meet(resolve.substitute(f.type)))
        return tuple(resolve.substitute(result.type) for result in node.results)

    # NOTE: unlike concrete interpreter, instead of using type information
    # within the IR. Type inference will use the interpreted
    # value (which is a type) to determine the method dispatch.
    def build_signature(
        self, frame: ForwardFrame[types.TypeAttribute], node: ir.Statement
    ) -> interp.Signature:
        argtypes = ()
        for x in frame.get_values(node.args):
            if isinstance(x, types.Generic):
                argtypes += (x.body,)
            else:
                argtypes += (x,)
        return interp.Signature(type(node), argtypes)



from __future__ import annotations

from abc import ABC
from typing import TypeVar, TypeAlias, overload
from dataclasses import field, dataclass

from kirin import ir
from kirin.lattice import BoundedLattice
from kirin.worklist import WorkList

from .abc import InterpreterABC
from .frame import Frame
from .value import Successor, YieldValue, ReturnValue
from .exceptions import InterpreterError

ResultType = TypeVar("ResultType", bound=BoundedLattice)
WorkListType = TypeVar("WorkListType", bound=WorkList[Successor])
AbsIntResultType: TypeAlias = (
    tuple[ResultType, ...] | None | ReturnValue[ResultType] | YieldValue[ResultType]
)


@dataclass
class AbstractFrame(Frame[ResultType]):
    """Interpreter frame for abstract interpreter.

    This frame is used to store the state of the abstract interpreter.
    It contains the worklist of successors to be processed.
    """

    worklist: WorkList[Successor[ResultType]] = field(default_factory=WorkList)
    visited: dict[ir.Block, set[Successor[ResultType]]] = field(default_factory=dict)


AbstractFrameType = TypeVar("AbstractFrameType", bound=AbstractFrame)


@dataclass
class AbstractInterpreter(InterpreterABC[AbstractFrameType, ResultType], ABC):
    """Abstract interpreter for the IR.

    This is a base class for implementing abstract interpreters for the IR.
    It provides a framework for implementing abstract interpreters given a
    bounded lattice type.

    The abstract interpreter is a forward dataflow analysis that computes
    the abstract values for each SSA value in the IR. The abstract values
    are computed by evaluating the statements in the IR using the abstract
    lattice operations.

    The abstract interpreter is implemented as a worklist algorithm. The
    worklist contains the successors of the current block to be processed.
    The abstract interpreter processes each successor by evaluating the
    statements in the block and updating the abstract values in the frame.

    The abstract interpreter provides hooks for customizing the behavior of
    the interpreter.
    The [`prehook_succ`][kirin.interp.abstract.AbstractInterpreter.prehook_succ] and
    [`posthook_succ`][kirin.interp.abstract.AbstractInterpreter.posthook_succ] methods
    can be used to perform custom actions before and after processing a successor.
    """

    lattice: type[BoundedLattice[ResultType]] = field(init=False)
    """lattice type for the abstract interpreter.
    """

    def __init_subclass__(cls) -> None:
        if ABC in cls.__bases__:
            return super().__init_subclass__()

        if not hasattr(cls, "lattice"):
            raise TypeError(
                f"missing lattice attribute in abstract interpreter class {cls}"
            )
        cls.void = cls.lattice.bottom()
        cls.keys += ("abstract",)
        super().__init_subclass__()

    def recursion_limit_reached(self) -> ResultType:
        return self.lattice.bottom()

    # helper methods
    @overload
    @staticmethod
    def join_results(old: None, new: None) -> None: ...
    @overload
    @staticmethod
    def join_results(
        old: ReturnValue[ResultType], new: ReturnValue[ResultType]
    ) -> ReturnValue[ResultType]: ...
    @overload
    @staticmethod
    def join_results(
        old: YieldValue[ResultType], new: YieldValue[ResultType]
    ) -> YieldValue[ResultType]: ...
    @overload
    @staticmethod
    def join_results(
        old: tuple[ResultType], new: tuple[ResultType]
    ) -> tuple[ResultType]: ...
    @overload
    @staticmethod
    def join_results(
        old: AbsIntResultType[ResultType], new: AbsIntResultType[ResultType]
    ) -> AbsIntResultType[ResultType]: ...

    @staticmethod
    def join_results(
        old: AbsIntResultType[ResultType],
        new: AbsIntResultType[ResultType],
    ) -> AbsIntResultType[ResultType]:
        if old is None:
            return new
        elif new is None:
            return old

        if isinstance(old, ReturnValue) and isinstance(new, ReturnValue):
            return ReturnValue(old.value.join(new.value))
        elif isinstance(old, YieldValue) and isinstance(new, YieldValue):
            return YieldValue(
                tuple(
                    old_val.join(new_val)
                    for old_val, new_val in zip(old.values, new.values)
                )
            )
        elif isinstance(old, tuple) and isinstance(new, tuple):
            return tuple(old_val.join(new_val) for old_val, new_val in zip(old, new))
        else:
            return None

    T = TypeVar("T")

    @classmethod
    def maybe_const(cls, value: ir.SSAValue, type_: type[T]) -> T | None:
        """Get a constant value of a given type.

        If the value is not a constant or the constant is not of the given type, return
        `None`.
        """
        from kirin.analysis.const.lattice import Value

        hint = value.hints.get("const")
        if isinstance(hint, Value) and isinstance(hint.data, type_):
            return hint.data

    @classmethod
    def expect_const(cls, value: ir.SSAValue, type_: type[T]):
        """Expect a constant value of a given type.

        If the value is not a constant or the constant is not of the given type, raise
        an `InterpreterError`.
        """
        hint = cls.maybe_const(value, type_)
        if hint is None:
            raise InterpreterError(f"expected {type_}, got {hint}")
        return hint



from kirin.serialization.base.context import MethodSymbolMeta
from kirin.serialization.base.serializationunit import SerializationUnit


class SerializationModule:
    symbol_table: dict[str, MethodSymbolMeta]
    body: SerializationUnit

    def __init__(
        self, symbol_table: dict[str, MethodSymbolMeta], body: SerializationUnit
    ):
        self.symbol_table = symbol_table
        self.body = body



from dataclasses import dataclass

from kirin import ir
from kirin.dialects import cf
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.analysis.cfg import CFG
from kirin.rewrite.walk import Walk
from kirin.rewrite.chain import Chain
from kirin.rewrite.fixpoint import Fixpoint


@dataclass
class DeadBlock(RewriteRule):
    """Compactify the CFG by removing dead blocks."""

    cfg: CFG

    def rewrite_Region(self, node: ir.Region) -> RewriteResult:
        # remove non-entry blocks that are not reachable from the entry block
        # TODO: check if this region is using SSACFG convention?
        has_done_something = False
        for block in node.blocks[1:]:
            predecessors = self.cfg.predecessors.get(block)
            if not predecessors:  # empty predecessors
                successors = self.cfg.successors.get(block, set())
                for successor in successors:
                    self.cfg.predecessors[successor].discard(block)
                self.cfg.successors.pop(block, None)
                self.cfg.predecessors.pop(block, None)
                block.delete()
                has_done_something = True
        return RewriteResult(has_done_something=has_done_something)


@dataclass
class CFGEdge(RewriteRule):
    """Merge non-branching blocks on the edge of the CFG.

    Example:

        /---> [B] --> [D] --> [E]
    [A]-----> [C] -------------^

    [B] and [D] are non-branching blocks on the same edge. They can be merged into one block.

        /---> [B,D] --> [E]
    [A]-----> [C] -------^
    """

    cfg: CFG

    def rewrite_Region(self, node: ir.Region) -> RewriteResult:
        result = RewriteResult()
        for block in node.blocks:
            result = self.rewrite_Block(block).join(result)
        return result

    def rewrite_Block(self, node: ir.Block) -> RewriteResult:
        successors = self.cfg.successors.get(node, None)
        if (
            successors is None or len(successors) > 1 or len(successors) == 0
        ):  # multiple outgoing edges
            return RewriteResult()

        successor = next(iter(successors))
        if len(self.cfg.predecessors[successor]) > 1:  # multiple incoming edges
            return RewriteResult()

        if not ((last_stmt := node.last_stmt) and isinstance(last_stmt, cf.Branch)):
            return RewriteResult()

        # merge the two blocks
        for arg, input in zip(successor.args, last_stmt.arguments):
            arg.replace_by(input)
        last_stmt.delete()
        for stmt in successor.stmts:
            stmt.detach()
            node.stmts.append(stmt)
        successor.delete()

        # update the CFG
        new_successors = self.cfg.successors[successor]
        self.cfg.successors[node] = new_successors
        for new_successor in new_successors:
            self.cfg.predecessors[new_successor].discard(successor)
            self.cfg.predecessors[new_successor].add(node)
        del self.cfg.successors[successor]
        del self.cfg.predecessors[successor]  # this is just [node]
        return RewriteResult(has_done_something=True)


class DuplicatedBranch(RewriteRule):
    """Merge duplicated branches into a single branch.

    Example:

        [A]-->[B]
          -----^

    Merge the two branches into one without changing the CFG:

        [A]-->[B]
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if (
            not isinstance(node, cf.ConditionalBranch)
            or node.then_successor is not node.else_successor
        ):
            return RewriteResult()

        for then_x, else_x in zip(node.then_arguments, node.else_arguments):
            if then_x is not else_x:
                return RewriteResult()

        node.replace_by(
            cf.Branch(arguments=node.then_arguments, successor=node.then_successor)
        )
        return RewriteResult(has_done_something=True)


@dataclass
class SkipBlock(RewriteRule):
    """Simplify a block that only contains a branch statement."""

    cfg: CFG

    def rewrite_Region(self, node: ir.Region) -> RewriteResult:
        result = RewriteResult()
        for block in node.blocks:
            result = self.rewrite_Block(block).join(result)
        return result

    def rewrite_Block(self, node: ir.Block) -> RewriteResult:
        if len(node.stmts) != 1:
            return RewriteResult()

        stmt = node.last_stmt
        if not isinstance(stmt, cf.Branch):
            return RewriteResult()

        has_done_something = False
        predecessors = self.cfg.predecessors.get(node, set())
        # only if there is one predecessor and no uses of the arguments
        if len(predecessors) == 1 and all(
            self.can_skip(stmt, each) for each in node.args
        ):
            has_done_something = self.rewrite_pred(node, stmt, next(iter(predecessors)))
        return RewriteResult(has_done_something=has_done_something)

    def can_skip(self, terminator: cf.Branch, value: ir.SSAValue) -> bool:
        for use in value.uses:
            if use.stmt is terminator:
                continue
            return False
        return True

    def rewrite_pred(
        self, node: ir.Block, node_terminator: cf.Branch, predecessor: ir.Block
    ) -> bool:
        terminator = predecessor.last_stmt
        if isinstance(terminator, cf.Branch):
            return self.rewrite_pred_Branch(
                node, node_terminator, predecessor, terminator
            )
        elif isinstance(terminator, cf.ConditionalBranch):
            return self.rewrite_pred_ConditionalBranch(
                node, node_terminator, predecessor, terminator
            )
        return False

    def rewrite_pred_Branch(
        self,
        node: ir.Block,
        node_terminator: cf.Branch,
        predecessor: ir.Block,
        pred_terminator: cf.Branch,
    ) -> bool:
        ssamap = self._block_inputs(node, pred_terminator.arguments)
        pred_terminator.replace_by(
            cf.Branch(
                # NOTE: the argument can also be SSAs from previous blocks (non-phi)
                arguments=tuple(
                    ssamap.get(arg, arg) for arg in node_terminator.arguments
                ),
                successor=node_terminator.successor,
            )
        )

        self.fix_cfg(predecessor, node, node_terminator.successor)
        return True

    def rewrite_pred_ConditionalBranch(
        self,
        node: ir.Block,
        node_terminator: cf.Branch,
        predecessor: ir.Block,
        pred_terminator: cf.ConditionalBranch,
    ) -> bool:
        then_arguments = pred_terminator.then_arguments
        else_arguments = pred_terminator.else_arguments
        then_successor = pred_terminator.then_successor
        else_successor = pred_terminator.else_successor

        has_done_something = False
        if pred_terminator.then_successor is node:
            ssamap = self._block_inputs(node, pred_terminator.then_arguments)
            then_arguments = tuple(
                ssamap.get(arg, arg) for arg in node_terminator.arguments
            )
            then_successor = node_terminator.successor
            has_done_something = True
            self.fix_cfg(predecessor, node, then_successor)

        if pred_terminator.else_successor is node:
            ssamap = self._block_inputs(node, pred_terminator.else_arguments)
            else_arguments = tuple(
                ssamap.get(arg, arg) for arg in node_terminator.arguments
            )
            else_successor = node_terminator.successor
            has_done_something = True
            self.fix_cfg(predecessor, node, else_successor)

        pred_terminator.replace_by(
            cf.ConditionalBranch(
                cond=pred_terminator.cond,
                then_arguments=then_arguments,
                then_successor=then_successor,
                else_arguments=else_arguments,
                else_successor=else_successor,
            )
        )
        return has_done_something

    def fix_cfg(self, predecessor: ir.Block, node: ir.Block, successor: ir.Block):
        node_pred_succ = self.cfg.successors.setdefault(predecessor, set())
        node_pred_succ.discard(node)
        node_pred_succ.add(successor)

        node_succ_pred = self.cfg.predecessors.setdefault(successor, set())
        node_succ_pred.add(predecessor)

        node_pred = self.cfg.predecessors.setdefault(node, set())
        node_pred.discard(predecessor)

    def _block_inputs(
        self, block: ir.Block, arguments: tuple[ir.SSAValue, ...]
    ) -> dict[ir.SSAValue, ir.SSAValue]:
        return dict(zip(block.args, arguments))


@dataclass
class CompactifyRegion(RewriteRule):
    """Wrapper to share the CFG object with same CFG region."""

    cfg: CFG

    def __init__(self, cfg: CFG):
        self.cfg = cfg
        self.rule = Fixpoint(
            Chain(
                DeadBlock(cfg), Walk(DuplicatedBranch()), SkipBlock(cfg), CFGEdge(cfg)
            )
        )

    def rewrite(self, node: ir.IRNode) -> RewriteResult:
        return self.rule.rewrite(node)


@dataclass
class CFGCompactify(RewriteRule):
    """Compactify the CFG by removing dead blocks and merging blocks
    if the statement uses the SSACFG convention. Do nothing if given
    `ir.Region` or `ir.Block` due to no context of the region.

    To compactify hierarchical CFG, combine this rule with `kirin.rewrite.Walk`
    to recursively apply this rule to all statements.
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        result = RewriteResult()
        if not (trait := node.get_trait(ir.HasCFG)):
            return result

        for region in node.regions:
            cfg = trait.get_graph(region)
            result = CompactifyRegion(cfg).rewrite(region).join(result)
        return result



"""The slice dialect for Python.

This dialect provides a `Slice` statement that represents a slice object in Python:

- The `Slice` statement class.
- The lowering pass for the `slice` call.
- The concrete implementation of the `slice` call.
- The type inference implementation of the `slice` call.
"""

import ast
from dataclasses import dataclass

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.print.printer import Printer
from kirin.dialects.py.constant import Constant

dialect = ir.Dialect("py.slice")


@dataclass(frozen=True)
class SliceLowering(lowering.FromPythonCall["Slice"]):

    def lower(
        self, stmt: type["Slice"], state: lowering.State, node: ast.Call
    ) -> lowering.Result:
        return _lower_slice(state, node)


T = types.TypeVar("T")


@statement(dialect=dialect, init=False)
class Slice(ir.Statement):
    name = "slice"
    traits = frozenset({ir.Pure(), SliceLowering()})
    start: ir.SSAValue = info.argument(T | types.NoneType)
    stop: ir.SSAValue = info.argument(T | types.NoneType)
    step: ir.SSAValue = info.argument(T | types.NoneType)
    result: ir.ResultValue = info.result(types.Slice[T])

    def __init__(
        self, start: ir.SSAValue, stop: ir.SSAValue, step: ir.SSAValue
    ) -> None:
        if not (
            isinstance(stop.type, types.TypeAttribute)
            and isinstance(start.type, types.TypeAttribute)
        ):
            result_type = types.Bottom
        elif start.type.is_subseteq(types.NoneType):
            if stop.type.is_subseteq(types.NoneType):
                result_type = types.Bottom
            else:
                result_type = types.Slice[stop.type]
        else:
            result_type = types.Slice[start.type]

        super().__init__(
            args=(start, stop, step),
            result_types=[result_type],
            args_slice={"start": 0, "stop": 1, "step": 2},
        )


@dataclass
class SliceAttribute(ir.Data[slice]):

    start: int | None
    stop: int | None
    step: int | None

    def __post_init__(self) -> None:
        if self.start is None and self.step is None:
            self.type = types.Slice[types.Literal(self.stop)]
        else:
            self.type = types.Slice3[
                types.Literal(self.start),
                types.Literal(self.stop),
                types.Literal(self.step),
            ]

    def unwrap(self):
        return slice(self.start, self.stop, self.step)

    def __hash__(self):
        return hash((type(self), self.start, self.stop, self.step))

    def print_impl(self, printer: Printer) -> None:
        return printer.plain_print(f"slice({self.start}, {self.stop}, {self.step})")

    def is_structurally_equal(
        self, other: ir.Attribute, context: dict | None = None
    ) -> bool:
        return (
            isinstance(other, SliceAttribute)
            and self.start == other.start
            and self.stop == other.stop
            and self.step == other.step
        )


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Slice)
    def _slice(self, interp, frame: interp.Frame, stmt: Slice):
        start, stop, step = frame.get_values(stmt.args)
        if start is None and step is None:
            return (SliceAttribute(None, stop, None),)
        elif step is None:
            return (SliceAttribute(start, stop, None),)
        else:
            return (SliceAttribute(start, stop, step),)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Slice(self, state: lowering.State, node: ast.Slice) -> lowering.Result:
        def value_or_none(expr: ast.expr | None) -> ir.SSAValue:
            if expr is not None:
                return state.lower(expr).expect_one()
            else:
                return state.current_frame.push(Constant(None)).result

        lower = value_or_none(node.lower)
        upper = value_or_none(node.upper)
        step = value_or_none(node.step)

        return state.current_frame.push(Slice(start=lower, stop=upper, step=step))

    @lowering.akin(slice)
    def lower_Call_slice(
        self, state: lowering.State, node: ast.Call
    ) -> lowering.Result:
        return _lower_slice(state, node)


def _lower_slice(state: lowering.State, node: ast.Call) -> lowering.Result:
    if len(node.args) == 1:
        start = state.lower(ast.Constant(None)).expect_one()
        stop = state.lower(node.args[0]).expect_one()
        step = state.lower(ast.Constant(None)).expect_one()
    elif len(node.args) == 2:
        start = state.lower(node.args[0]).expect_one()
        stop = state.lower(node.args[1]).expect_one()
        step = state.lower(ast.Constant(None)).expect_one()
    elif len(node.args) == 3:
        start = state.lower(node.args[0]).expect_one()
        stop = state.lower(node.args[1]).expect_one()
        step = state.lower(node.args[2]).expect_one()
    else:
        raise lowering.BuildError("slice() takes 1-3 arguments")

    return state.current_frame.push(Slice(start, stop, step))



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar
from dataclasses import dataclass

from kirin.ir.attrs.py import PyAttr
from kirin.ir.exception import ValidationError
from kirin.ir.traits.abc import StmtTrait

if TYPE_CHECKING:
    from kirin.ir import Statement


@dataclass(frozen=True)
class SymbolOpInterface(StmtTrait):
    """A trait that indicates that a statement is a symbol operation.

    A symbol operation is a statement that has a symbol name attribute.
    """

    def get_sym_name(self, stmt: Statement) -> PyAttr[str]:
        sym_name: PyAttr[str] | None = stmt.get_attribute("sym_name")  # type: ignore
        # NOTE: unlike MLIR or xDSL we do not allow empty symbol names
        if sym_name is None:
            raise ValueError(f"Statement {stmt.name} does not have a symbol name")
        return sym_name

    def verify(self, node: Statement):
        from kirin.types import String

        sym_name = self.get_sym_name(node)
        if not (isinstance(sym_name, PyAttr) and sym_name.type.is_subseteq(String)):
            raise ValueError(f"Symbol name {sym_name} is not a string attribute")


@dataclass(frozen=True)
class SymbolTable(StmtTrait):
    """
    Statement with SymbolTable trait can only have one region with one block.
    """

    @staticmethod
    def walk(stmt: Statement):
        for stmt in stmt.regions[0].blocks[0].stmts:
            if stmt.has_trait(SymbolOpInterface):
                yield stmt

    def verify(self, node: Statement):
        if len(node.regions) != 1:
            raise ValidationError(
                node,
                f"Statement {node.name} with SymbolTable trait must have exactly one region",
            )

        if len(node.regions[0].blocks) != 1:
            raise ValidationError(
                node,
                f"Statement {node.name} with SymbolTable trait must have exactly one block",
            )

        # TODO: check uniqueness of symbol names


StmtType = TypeVar("StmtType", bound="Statement")


@dataclass(frozen=True)
class EntryPointInterface(StmtTrait, Generic[StmtType]):
    """A trait that indicates that a module-like statement has an entry point.

    An entry point is a statement that has a symbol name attribute and is
    the first statement in the module.

    When interpreting statements with this trait, the interpreter will
    look for the entry point and start calling the module from there.
    """

    @abstractmethod
    def get_entry_point_symbol(self, stmt: StmtType) -> str: ...

    @abstractmethod
    def get_entry_point(self, stmt: StmtType) -> Statement: ...

    def verify(self, node: Statement):
        if not node.has_trait(SymbolOpInterface):
            raise ValidationError(
                node,
                f"Statement {node.name} with EntryPointInterface trait must have a symbol name",
            )



from __future__ import annotations

from typing import IO, Generic, TypeVar
from contextlib import contextmanager
from dataclasses import field, dataclass

from kirin import ir, interp
from kirin.idtable import IdTable
from kirin.worklist import WorkList

from .abc import EmitABC, EmitFrame

IO_t = TypeVar("IO_t", bound=IO)


@dataclass
class SymbolTable(IdTable[ir.Statement]):

    def add(self, value: ir.Statement) -> str:
        id = self.next_id
        if (trait := value.get_trait(ir.SymbolOpInterface)) is not None:
            value_name = trait.get_sym_name(value).unwrap()
            curr_ind = self.name_count.get(value_name, 0)
            suffix = f"_{curr_ind}" if curr_ind != 0 else ""
            self.name_count[value_name] = curr_ind + 1
            name = self.prefix + value_name + suffix
            self.table[value] = name
        else:
            name = f"{self.prefix}{self.prefix_if_none}{id}"
            self.next_id += 1
            self.table[value] = name
        return name

    def __getitem__(self, value: ir.Statement) -> str:
        if value in self.table:
            return self.table[value]
        raise KeyError(f"Symbol {value} not found in SymbolTable")

    def get(self, value: ir.Statement, default: str | None = None) -> str | None:
        if value in self.table:
            return self.table[value]
        return default


@dataclass
class JuliaFrame(EmitFrame[str], Generic[IO_t]):
    io: IO_t
    ssa: IdTable[ir.SSAValue] = field(
        default_factory=lambda: IdTable[ir.SSAValue](prefix="ssa_")
    )
    block: IdTable[ir.Block] = field(
        default_factory=lambda: IdTable[ir.Block](prefix="block_")
    )
    _indent: int = 0

    def write(self, value):
        self.io.write(value)

    def write_line(self, value):
        self.write("    " * self._indent + value + "\n")

    @contextmanager
    def indent(self):
        self._indent += 1
        yield
        self._indent -= 1


@dataclass
class Julia(EmitABC[JuliaFrame, str], Generic[IO_t]):
    """Julia code generator for the IR.

    This class generates Julia code from the IR.
    It is used to generate Julia code for the IR.
    """

    keys = ("emit.julia",)
    void = ""

    # some states
    io: IO_t
    callables: SymbolTable = field(init=False)
    callable_to_emit: WorkList[ir.Statement] = field(init=False)

    def initialize(self):
        super().initialize()
        self.callables = SymbolTable(prefix="_callable_")
        self.callable_to_emit = WorkList()
        return self

    def initialize_frame(
        self, node: ir.Statement, *, has_parent_access: bool = False
    ) -> JuliaFrame:
        return JuliaFrame(node, self.io, has_parent_access=has_parent_access)

    def run(self, node: ir.Method | ir.Statement):
        if isinstance(node, ir.Method):
            node = node.code

        with self.eval_context():
            self.callables.add(node)
            self.callable_to_emit.append(node)
            while self.callable_to_emit:
                callable = self.callable_to_emit.pop()
                if callable is None:
                    break
                self.eval(callable)
                self.io.flush()
        return

    def frame_call(
        self, frame: JuliaFrame, node: ir.Statement, *args: str, **kwargs: str
    ) -> str:
        return f"{args[0]}({', '.join(args[1:])})"

    def get_attribute(self, frame: JuliaFrame, node: ir.Attribute) -> str:
        method = self.registry.get(interp.Signature(type(node)))
        if method is None:
            raise ValueError(f"Method not found for node: {node}")
        return method(self, frame, node)



from kirin import types
from kirin.prelude import basic
from kirin.dialects.ilist import IList, IListType


@basic(typeinfer=True)
def tuple_new(x: int, xs: tuple):
    return xs + (1, x)


@basic(typeinfer=True)
def list_new(x: int, xs: IList):
    return xs + [1, x]


def test_tuple_add():
    assert tuple_new.return_type.is_subseteq(types.Tuple)
    assert list_new.return_type.is_subseteq(IListType)



from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk, Chain, Fixpoint, WrapConst
from kirin.analysis import const
from kirin.rewrite.dce import DeadCodeElimination
from kirin.rewrite.getitem import InlineGetItem


@basic_no_opt
def main_simplify_getitem(x: int):
    ylist = (x, x, 1, 2)
    return ylist[0]


def test_getitem():
    before = main_simplify_getitem(1)
    constprop = const.Propagate(main_simplify_getitem.dialects)
    frame, _ = constprop.run(main_simplify_getitem)
    Fixpoint(Walk(WrapConst(frame))).rewrite(main_simplify_getitem.code)
    inline_getitem = InlineGetItem()
    Fixpoint(Walk(Chain([inline_getitem, DeadCodeElimination()]))).rewrite(
        main_simplify_getitem.code
    )
    main_simplify_getitem.code.print()
    after = main_simplify_getitem(1)
    assert before == after
    assert len(main_simplify_getitem.callable_region.blocks[0].stmts) == 1



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar, Iterable
from dataclasses import dataclass

from kirin import ir, interp, lattice

LatticeType = TypeVar("LatticeType", bound=lattice.BoundedLattice)


@dataclass
class ForwardFrame(interp.AbstractFrame[LatticeType]):

    def set_values(
        self, keys: Iterable[ir.SSAValue], values: Iterable[LatticeType]
    ) -> None:
        for ssa_value, result in zip(keys, values):
            if ssa_value in self.entries:
                self.entries[ssa_value] = self.entries[ssa_value].join(result)
            else:
                self.entries[ssa_value] = result


FrameType = TypeVar("FrameType", bound=ForwardFrame)


@dataclass
class ForwardExtra(interp.AbstractInterpreter[FrameType, LatticeType], ABC):
    """Forward dataflow analysis but with custom frame for extra information/state
    per call frame.

    Params:
        FrameType: The type of the frame used for the analysis.
        LatticeType: The type of the lattice used for the analysis.
    """

    def run(self, method: ir.Method, *args: LatticeType, **kwargs: LatticeType):
        if not args and not kwargs:  # empty args and kwargs
            args = tuple(self.lattice.top() for _ in method.args)
        return self.call(method, self.method_self(method), *args, **kwargs)

    def run_no_raise(
        self, method: ir.Method, *args: LatticeType, **kwargs: LatticeType
    ):
        try:
            return self.run(method, *args, **kwargs)
        except Exception:
            return self.initialize_frame(method.code), self.lattice.bottom()

    @abstractmethod
    def method_self(self, method: ir.Method) -> LatticeType:
        """Return the self value for the given method."""
        ...


@dataclass
class Forward(ForwardExtra[ForwardFrame[LatticeType], LatticeType], ABC):
    """Forward dataflow analysis.

    This is the base class for forward dataflow analysis. If your analysis
    requires extra information per frame, you should subclass
    [`ForwardExtra`][kirin.analysis.forward.ForwardExtra] instead.
    """

    def initialize_frame(
        self, node: ir.Statement, *, has_parent_access: bool = False
    ) -> ForwardFrame[LatticeType]:
        return ForwardFrame(node, has_parent_access=has_parent_access)



from dataclasses import dataclass

from kirin import ir, types
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.func.attrs import Signature


@dataclass
class ApplyType(RewriteRule):
    results: dict[ir.SSAValue, types.TypeAttribute]

    def rewrite_Block(self, node: ir.Block) -> RewriteResult:
        has_done_something = False
        for arg in node.args:
            if arg in self.results:
                arg.type = self.results[arg]
                has_done_something = True

        return RewriteResult(has_done_something=has_done_something)

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        has_done_something = False
        for result in node._results:
            if result in self.results:
                result.type = self.results[result]
                has_done_something = True

        if (trait := node.get_trait(ir.HasSignature)) is not None and (
            callable_trait := node.get_trait(ir.CallableStmtInterface)
        ) is not None:
            callable_region = callable_trait.get_callable_region(node)
            inputs = tuple(
                self.results.get(arg, arg.type)
                for arg in callable_region.blocks[0].args
            )

            if (
                len(node._results) == 1
                and isinstance(
                    output_ := self.results.get(node._results[0]), types.Generic
                )
                and output_.is_subseteq(types.MethodType)
            ):
                output_ = output_.vars[1]
                trait.set_signature(node, Signature(inputs, output_))
                has_done_something = True
        return RewriteResult(has_done_something=has_done_something)



import pytest

from kirin import ir
from kirin.prelude import structural_no_opt


def test_python_func():
    def some_func(x):
        return x + 1

    @structural_no_opt
    def dumm(x):
        return some_func(x)

    with pytest.raises(ir.TypeCheckError):
        dumm.code.verify_type()

    some_staff = ""

    @structural_no_opt
    def dumm2(x):
        return some_staff(x)  # type: ignore

    with pytest.raises(ir.TypeCheckError):
        dumm.code.verify_type()



from kirin import emit, interp

from .stmts import Add
from ._dialect import dialect


@dialect.register(key="emit.julia")
class JuliaEmit(interp.MethodTable):
    @interp.impl(Add)
    def add(self, emit_: emit.Julia, frame: emit.JuliaFrame, node: Add):
        lhs = frame.get(node.lhs)
        rhs = frame.get(node.rhs)
        frame.write_line(f"{frame.ssa[node.result]} = ({lhs} + {rhs})")
        return (frame.ssa[node.result],)



from __future__ import annotations

from typing import Generic, TypeVar, Iterable
from dataclasses import field, dataclass

ElemType = TypeVar("ElemType")


@dataclass
class WorkList(Generic[ElemType]):
    """The worklist data structure.

    The worklist is a stack that allows for O(1) removal of elements from the stack.
    """

    _stack: list[ElemType] = field(default_factory=list, init=False)

    def __len__(self) -> int:
        return len(self._stack)

    def __bool__(self) -> bool:
        return bool(self._stack)

    def append(self, item: ElemType) -> None:
        self._stack.append(item)

    def extend(self, items: Iterable[ElemType]) -> None:
        self._stack.extend(items)

    def pop(self) -> ElemType | None:
        if self._stack:
            return self._stack.pop(0)
        return None



from typing import Any, Literal

from kirin import ir, types, rewrite
from kirin.passes import aggressive
from kirin.prelude import basic_no_opt, python_basic
from kirin.analysis import const
from kirin.dialects import py, func, ilist, lowering
from kirin.passes.typeinfer import TypeInfer


@ir.dialect_group(
    python_basic.union([func, ilist, lowering.func, lowering.range.ilist])
)
def basic(self):
    aggressive_fold_pass = aggressive.Fold(self)
    typeinfer_pass = TypeInfer(self)

    def run_pass(
        mt: ir.Method,
    ) -> None:
        aggressive_fold_pass.fixpoint(mt)
        rewrite.Fixpoint(rewrite.Walk(ilist.rewrite.ConstList2IList())).rewrite(mt.code)
        typeinfer_pass(mt)

    return run_pass


def test_empty():
    @basic
    def empty_list():
        return []

    empty_list.print()
    assert empty_list.return_type.is_subseteq(ilist.IListType[types.Any])


def test_typehint():
    @basic
    def main(xs: ilist.IList[int, Literal[3]]):
        return xs + [4, 5, 6] + xs

    assert main.return_type is not None
    assert main.return_type.is_subseteq(ilist.IListType[types.Int, types.Literal(9)])


@basic
def add1(x: int):
    return x + 1


def test_ilist_fcf():
    # TODO: actually check equivalent code
    rule = rewrite.Fixpoint(rewrite.Walk(ilist.rewrite.Unroll()))

    xs = ilist.IList([1, 2, 3])

    @basic
    def map(xs: ilist.IList[int, Literal[3]]):
        return ilist.map(add1, xs)

    @basic_no_opt
    def foreach(xs: ilist.IList[int, Literal[3]]):
        ilist.for_each(add1, xs)

    map_before = map(xs)
    foreach_before = foreach(xs)
    rule.rewrite(map.code)
    rule.rewrite(foreach.code)
    map_after = map(xs)
    foreach_after = foreach(xs)
    assert map_before.data == map_after.data  # type: ignore
    assert foreach_before == foreach_after

    assert isinstance(map.callable_region.blocks[0].stmts.at(1), py.Constant)
    assert isinstance(map.callable_region.blocks[0].stmts.at(-2), ilist.New)

    assert isinstance(foreach.callable_region.blocks[0].stmts.at(1), py.Constant)
    assert isinstance(
        foreach.callable_region.blocks[0].stmts.at(2), py.indexing.GetItem
    )
    assert isinstance(foreach.callable_region.blocks[0].stmts.at(3), func.Call)
    assert isinstance(foreach.callable_region.blocks[0].stmts.at(10), func.ConstantNone)

    @basic
    def add(x: int, y: int):
        return x + y, y

    @basic
    def scan(xs: ilist.IList[int, Literal[3]]):
        return ilist.Scan(add, xs, init=123)  # type: ignore

    scan_before = scan(xs)
    rule.rewrite(scan.code)
    scan_after = scan(xs)
    assert scan_before == scan_after  # type: ignore
    assert isinstance(scan.callable_region.blocks[0].stmts.at(-2), py.tuple.New)
    assert isinstance(scan.callable_region.blocks[0].stmts.at(-3), ilist.New)

    @basic
    def add2(x: int, y: int):
        return x + y

    @basic
    def foldl(xs: ilist.IList[int, Literal[3]]):
        return ilist.Foldl(add2, xs, init=123)  # type: ignore

    @basic
    def foldr(xs: ilist.IList[int, Literal[3]]):
        return ilist.Foldr(add2, xs, init=123)  # type: ignore

    foldl_before = foldl(xs)
    foldr_before = foldr(xs)
    rule.rewrite(foldl.code)
    rule.rewrite(foldr.code)
    foldl_after = foldl(xs)
    foldr_after = foldr(xs)

    assert foldl_before == foldl_after  # type: ignore
    assert foldr_before == foldr_after  # type: ignore

    stmt = foldl.callable_region.blocks[0].stmts.at(2)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == 0
    assert isinstance(foldl.callable_region.blocks[0].stmts.at(4), func.Call)

    stmt = foldl.callable_region.blocks[0].stmts.at(5)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == 1
    assert isinstance(foldl.callable_region.blocks[0].stmts.at(7), func.Call)

    stmt = foldl.callable_region.blocks[0].stmts.at(8)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == 2
    assert isinstance(foldl.callable_region.blocks[0].stmts.at(10), func.Call)

    # ========== foldl
    stmt = foldr.callable_region.blocks[0].stmts.at(2)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == 2
    assert isinstance(foldr.callable_region.blocks[0].stmts.at(4), func.Call)

    stmt = foldr.callable_region.blocks[0].stmts.at(5)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == 1
    assert isinstance(foldr.callable_region.blocks[0].stmts.at(7), func.Call)

    stmt = foldr.callable_region.blocks[0].stmts.at(8)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == 0
    assert isinstance(foldr.callable_region.blocks[0].stmts.at(10), func.Call)


def test_ilist_range():
    @basic.add(py.range)
    def map():
        return ilist.Map(add1, range(0, 3))  # type: ignore

    assert map() == ilist.IList([1, 2, 3])

    @basic.add(py.range)
    def const_range():
        return range(0, 3)

    assert const_range() == ilist.IList(range(0, 3))


def test_inline_get_item_integer():
    items = tuple(ir.TestValue() for _ in range(2))

    test_block = ir.Block(
        [
            qreg := ilist.New(values=items),
            idx := py.Constant(0),
            qubit_stmt := py.GetItem(obj=qreg.result, index=idx.result),
            ilist.New(values=(qubit_stmt.result,)),
            idx1 := py.Constant(10),
            qubit_stmt := py.GetItem(obj=qreg.result, index=idx1.result),
            ilist.New(values=(qubit_stmt.result,)),
        ]
    )

    idx.result.hints["const"] = const.Value(0)
    idx1.result.hints["const"] = const.Value(10)
    rule = rewrite.Walk(ilist.rewrite.InlineGetItem())
    rule.rewrite(test_block)

    expected_block = ir.Block(
        [
            qreg := ilist.New(values=items),
            idx := py.Constant(0),
            qubit_stmt := py.GetItem(obj=qreg.result, index=idx.result),
            ilist.New(values=(items[0],)),
            idx1 := py.Constant(10),
            qubit_stmt := py.GetItem(obj=qreg.result, index=idx1.result),
            ilist.New(values=(qubit_stmt.result,)),
        ]
    )

    assert test_block.is_structurally_equal(expected_block)


def test_inline_getitem_slice():
    values = tuple(ir.TestValue() for _ in range(6))
    test_block = ir.Block(
        [
            qreg := ilist.New(values=values),
            slice_value := py.Constant(slice(2, 5, 1)),
            py.GetItem(obj=qreg.result, index=slice_value.result),
        ]
    )
    slice_value.result.hints["const"] = const.Value(slice(2, 5, 1))

    rule = rewrite.Walk(ilist.rewrite.InlineGetItem())
    rule.rewrite(test_block)

    expected_block = ir.Block(
        [
            qreg := ilist.New(values=values),
            slice_value := py.Constant(slice(2, 5, 1)),
            ilist.New(values=(values[2], values[3], values[4])),
        ]
    )
    assert test_block.is_structurally_equal(expected_block)


def test_ilist_flatten_add_rhs_empty():
    values = tuple(ir.TestValue() for _ in range(6))
    test_block = ir.Block(
        [
            lhs := ilist.New(values=values),
            rhs := py.Constant(ilist.IList([])),
            py.Add(lhs=lhs.result, rhs=rhs.result),
        ]
    )
    rhs.result.hints["const"] = const.Value(ilist.IList([]))

    rule = rewrite.Walk(ilist.rewrite.FlattenAdd())
    rule.rewrite(test_block)

    expected_block = ir.Block(
        [
            lhs := ilist.New(values=values),
            rhs := py.Constant(ilist.IList([])),
            ilist.New(values=values),
        ]
    )
    assert test_block.is_structurally_equal(expected_block)


def test_ilist_flatten_add_lhs_empty():
    values = tuple(ir.TestValue() for _ in range(6))
    test_block = ir.Block(
        [
            lhs := py.Constant(ilist.IList([])),
            rhs := ilist.New(values=values),
            py.Add(lhs=lhs.result, rhs=rhs.result),
        ]
    )
    lhs.result.hints["const"] = const.Value(ilist.IList([]))

    rule = rewrite.Walk(ilist.rewrite.FlattenAdd())
    rule.rewrite(test_block)

    expected_block = ir.Block(
        [
            lhs := py.Constant(ilist.IList([])),
            rhs := ilist.New(values=values),
            ilist.New(values=values),
        ]
    )
    assert test_block.is_structurally_equal(expected_block)


def test_ilist_flatten_add_lhs_not_empty():
    values = tuple(ir.TestValue() for _ in range(6))
    test_block = ir.Block(
        [
            lhs := py.Constant(value := ilist.IList([1])),
            rhs := ilist.New(values=values),
            py.Add(lhs=lhs.result, rhs=rhs.result),
        ]
    )
    lhs.result.hints["const"] = const.Value(value)

    rule = rewrite.Walk(ilist.rewrite.FlattenAdd())
    result = rule.rewrite(test_block)

    assert not result.has_done_something


def test_ilist_flatten_add_rhs_not_empty():
    values = tuple(ir.TestValue() for _ in range(6))
    test_block = ir.Block(
        [
            lhs := ilist.New(values=values),
            rhs := py.Constant(value := ilist.IList([1])),
            py.Add(lhs=lhs.result, rhs=rhs.result),
        ]
    )
    rhs.result.hints["const"] = const.Value(value)

    rule = rewrite.Walk(ilist.rewrite.FlattenAdd())
    result = rule.rewrite(test_block)

    assert not result.has_done_something


def test_ilist_flatten_add_both_new():
    lhs_values = tuple(ir.TestValue() for _ in range(6))
    rhs_values = tuple(ir.TestValue() for _ in range(3))
    test_block = ir.Block(
        [
            lhs := ilist.New(values=lhs_values),
            rhs := ilist.New(values=rhs_values),
            py.Add(lhs=lhs.result, rhs=rhs.result),
        ]
    )
    expected_block = ir.Block(
        [
            lhs := ilist.New(values=lhs_values),
            rhs := ilist.New(values=rhs_values),
            ilist.New(values=lhs_values + rhs_values),
        ]
    )
    rule = rewrite.Walk(ilist.rewrite.FlattenAdd())
    rule.rewrite(test_block)

    assert test_block.is_structurally_equal(expected_block)


def test_ilist_constprop():
    from kirin.analysis import const

    @basic_no_opt
    def test_impl(x: float) -> float:
        return x * 0.3

    @basic_no_opt
    def _for_loop(
        values: ilist.IList[float, Any],
    ) -> ilist.IList[float, Any]:

        def gen(i: int):
            return test_impl(
                x=values[i],
            )

        n_range = len(values)
        return ilist.map(fn=gen, collection=ilist.range(n_range))

    @basic_no_opt
    def main():
        values = [1.0, 2.0, 3.0]
        return _for_loop(values)  # type: ignore

    prop = const.Propagate(basic_no_opt)
    frame, result = prop.run(main)
    target_ssa = main.callable_region.blocks[0].stmts.at(-2).results[0]
    target = frame.entries[target_ssa]
    assert isinstance(target, const.Value)
    for x, y in zip(target.data, ilist.IList([0.3, 0.6, 0.9])):
        assert abs(x - y) < 1e-9

    @basic_no_opt
    def add(x, y):
        return x + y

    @basic_no_opt
    def foldl(xs):
        return ilist.foldl(add, xs, init=0)

    @basic_no_opt
    def foldr(xs):
        return ilist.foldr(add, xs, init=0)

    @basic_no_opt
    def main2():
        values = [1, 2, 3]
        return foldl(values), foldr(values)

    prop = const.Propagate(basic_no_opt)
    frame, result = prop.run(main2)
    target_ssa = main2.callable_region.blocks[0].stmts.at(-2).results[0]
    target = frame.entries[target_ssa]
    assert isinstance(target, const.Value)
    assert target.data == (6, 6)


rule = rewrite.Fixpoint(rewrite.Walk(ilist.rewrite.Unroll()))
xs = ilist.IList([1, 2, 3])


@basic
def map(xs: ilist.IList[int, Literal[3]]):
    return ilist.map(add1, xs)


@basic_no_opt
def foreach(xs: ilist.IList[int, Literal[3]]):
    ilist.for_each(add1, xs)


map_before = map(xs)
foreach_before = foreach(xs)
rule.rewrite(map.code)
rule.rewrite(foreach.code)
map_after = map(xs)
foreach_after = foreach(xs)



from typing import Any
from dataclasses import MISSING

from typing_extensions import Unpack

from kirin import ir, types
from kirin.decl import info
from kirin.decl.base import BaseModifier, StatementOptions

from ._create_fn import create_fn
from ._set_new_attribute import set_new_attribute


class _HAS_DEFAULT_FACTORY_CLASS:
    def __repr__(self):
        return "<factory>"


_HAS_DEFAULT_FACTORY = _HAS_DEFAULT_FACTORY_CLASS()


class EmitInit(BaseModifier):
    _POST_INIT_NAME = "__post_init__"
    _RESULT_VALUE_NAME = "_kirin_ResultValue"
    _KIRIN_PYATTR = "_kirin_PyAttr"
    _ATTR_FACTORY_PREFIX = "_kirin_attr_factory_"
    _SELF_CLASS = "_kirin_self_class"
    _KIRIN_STMT = "_kirin_Statement"

    def __init__(self, cls: type, **kwargs: Unpack[StatementOptions]) -> None:
        super().__init__(cls, **kwargs)
        from kirin.ir.attrs.py import PyAttr

        self._init_params: list[str] = []
        self._init_body: list[str] = []
        self._init_locals: dict[str, Any] = {}
        self.has_post_init = hasattr(self.cls, self._POST_INIT_NAME)
        self.globals.update(
            {
                "_PY_ANY": types.Any,
                self._KIRIN_STMT: ir.Statement,
                self._SELF_CLASS: self.cls,
                self._RESULT_VALUE_NAME: ir.ResultValue,
                self._KIRIN_PYATTR: PyAttr,
            }
        )

    @classmethod
    def _hint_name(cls, f: info.Field):
        return f"_hint_{f.name}"

    def emit_init(self):
        _args_groups = []
        for f in self.fields.args.values():
            if f.group:
                _args_groups.append(f.name)
        setattr(self.cls, "_arg_groups", frozenset(_args_groups))

        if "__init__" in self.cls.__dict__:
            return True

        self._init_locals.update(
            {self._hint_name(f): f.annotation for f in self.fields}
        )
        self._init_locals.update(
            {
                "MISSING": MISSING,
                "_HAS_DEFAULT_FACTORY": _HAS_DEFAULT_FACTORY,
            }
        )

        self._emit_params()
        self._emit_field_init(len(_args_groups) > 0)

        fn = create_fn(
            "__init__",
            self._init_params,
            self._init_body,
            globals=self.globals,
            locals=self._init_locals,
            return_type=None,
        )
        set_new_attribute(self.cls, "__init__", fn)

    def _emit_params(self):
        self._init_params.append(self._self_name)
        for f in self.fields.std_args.values():
            self._init_params.append(self._init_param_value(f))

        kw_params: list[str] = []
        for f in self.fields.kw_args.values():
            kw_params.append(self._init_param_value(f))

        for f in self.fields.attributes.values():
            if f.init:
                kw_params.append(self._init_param_value(f))

        for f in self.fields.regions.values():
            if f.init:
                kw_params.append(self._init_param_value(f))

        for f in self.fields.blocks.values():
            if f.init:
                kw_params.append(self._init_param_value(f))

        if kw_params:
            self._init_params.append("*")
            self._init_params.extend(kw_params)

    def _emit_field_init(self, has_args_groups: bool):
        self._init_body.append(
            (
                f"{self._KIRIN_STMT}.__init__({self._self_name},"
                f"args={self._arg_seq()},"
                f"regions={self._regions_seq()},"
                f"successors={self._blocks_seq()},"
                f"result_types={self._result_types_seq()},"
                f"attributes={self._attribute_seq()},"
                f"args_slice={self._args_slice(has_args_groups)}"
                ")"
            )
        )

        # TODO: support InitVar?
        if self.has_post_init:
            self._init_body.append(f"{self._self_name}.{self._POST_INIT_NAME}()")

    def _init_param_value(self, f: info.Field):
        if isinstance(f, info.AttributeField):  # only attribute can have defaults
            if f.default is MISSING and f.default_factory is None:
                default = ""
            elif f.default is not MISSING:
                default = f"={self._ATTR_FACTORY_PREFIX}{f.name}"
            elif f.default_factory is not None:
                default = "=_HAS_DEFAULT_FACTORY"
            else:
                raise ValueError("unexpected default")
            return f"{f.name}:{self._hint_name(f)}{default}"
        elif isinstance(f, info.ArgumentField):  # arguments are always required
            return f"{f.name}:{self._hint_name(f)}"
        elif isinstance(f, info.ResultField):
            raise ValueError("result fields are not allowed in __init__")
        elif isinstance(f, info.RegionField):
            return f"{f.name}:{self._hint_name(f)}=_HAS_DEFAULT_FACTORY"
        elif isinstance(f, info.BlockField):
            return f"{f.name}:{self._hint_name(f)}=_HAS_DEFAULT_FACTORY"
        else:  # the rest fields are all required
            raise ValueError(f"unexpected field type {type(f)}")

    def _arg_seq(self):
        args: list[str] = [self._field_param(f) for f in self.fields.args.values()]
        return self._tuple_str(args)

    def _field_param(self, f: info.ArgumentField):
        if f.group:
            return f"*{f.name}"
        else:
            return f"{f.name}"

    def _result_types_seq(self):
        result_types = [
            self._result_type_value(f) for f in self.fields.results.values()
        ]
        return self._tuple_str(result_types)

    def _result_type_value(self, f: info.ResultField):
        name = f"_result_type_{f.name}"
        self._init_locals[name] = f.type
        return name

    def _attribute_seq(self):
        attrs = ",".join(
            [
                f'"{name}": {self._attribute_value(f)}'
                for name, f in self.fields.attributes.items()
            ]
        )
        return "{" + attrs + "}"

    def _attribute_value(self, f: info.AttributeField):
        default_name = f"{self._ATTR_FACTORY_PREFIX}{f.name}"
        if f.default_factory is not None:
            value = self._field_with_default_factory(
                default_name, f.name, f.init, f.default_factory
            )
        else:
            # no default factory
            if f.init:
                if f.default is not MISSING:
                    self._init_locals[default_name] = f.default
                value: str = f.name  # type: ignore
            else:  # no default factory, no init, no default
                raise ValueError(
                    "attribute must have a default or default factory or be initialized (init=True)"
                )

        # declared via python type, optionally check if we can
        # convert the value to data.PyAttr
        if f.pytype:
            attr_type = f"_kirin_attr_type_{f.name}"
            self._init_locals[attr_type] = f.type
            value = (
                f"{self._KIRIN_PYATTR}({value}, {attr_type}) "
                f"if not isinstance({value}, {self._KIRIN_PYATTR}) "
                f"else {value}"
            )

        return value

    def _regions_seq(self):
        regions = [self._regions_value(f) for f in self.fields.regions.values()]
        return self._tuple_str(regions)

    def _regions_value(self, f: info.RegionField):
        return self._field_with_default_factory(
            f"_kirin_region_{f.name}", f.name, f.init, f.default_factory
        )

    def _blocks_seq(self):
        blocks = [self._blocks_value(f) for f in self.fields.blocks.values()]
        return self._tuple_str(blocks)

    def _blocks_value(self, f: info.BlockField):
        return self._field_with_default_factory(
            f"_kirin_block_{f.name}", f.name, f.init, f.default_factory
        )

    def _args_slice(self, has_args_groups: bool):
        if not has_args_groups:
            return (
                "{"
                + ", ".join(
                    [
                        f"'{f.name}': {index}"
                        for index, f in enumerate(self.fields.args.values())
                    ]
                )
                + "}"
            )

        # NOTE: SSAValue fields do not have default/default_factory
        # so we can just count the input arguments
        ret: list[str] = []
        self._init_body.append("__args_slice_count = 0")
        for f in self.fields.args.values():
            if f.group:
                self._init_body.append(
                    f"__{f.name}_slice = slice(__args_slice_count, __args_slice_count + len({f.name}))"
                )
                self._init_body.append(f"__args_slice_count += len({f.name})")
            else:
                self._init_body.append(f"__{f.name}_slice = __args_slice_count")
                self._init_body.append("__args_slice_count += 1")
            ret.append(f'"{f.name}": __{f.name}_slice')
        return "{" + ",".join(ret) + "}"

    def _field_with_default_factory(
        self, factory_name: str, name: str | None, init: bool, default_factory
    ):
        # block always has default_factory
        if init:
            self._init_locals[factory_name] = default_factory
            value = (
                f"{factory_name}() "
                f"if {name} is _HAS_DEFAULT_FACTORY "
                f"else {name}"
            )
        else:
            self._init_locals[factory_name] = default_factory
            value = f"{factory_name}()"
        return value

    @staticmethod
    def _tuple_str(seq: list[str]):
        if not seq:
            return "()"
        return f"({', '.join(seq)}, )"



from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk, Fixpoint, WrapConst
from kirin.analysis import const
from kirin.rewrite.dce import DeadCodeElimination
from kirin.rewrite.fold import ConstantFold
from kirin.rewrite.compactify import CFGCompactify


@basic_no_opt
def branch(x):
    if x > 1:
        y = x + 1
    else:
        y = x + 2

    if True:
        return y + 1
    else:
        y + 2


def test_branch_elim():
    assert branch(1) == 4
    const_prop = const.Propagate(branch.dialects)
    frame, ret = const_prop.run(branch)
    Walk(Fixpoint(WrapConst(frame))).rewrite(branch.code)
    fold = ConstantFold()
    Fixpoint(Walk(fold)).rewrite(branch.code)
    # TODO: also check the generated CFG
    # interp.worklist.visited
    Fixpoint(CFGCompactify()).rewrite(branch.code)
    Walk(DeadCodeElimination()).rewrite(branch.code)
    branch.code.print()
    assert len(branch.code.body.blocks) == 4  # type: ignore



import pytest

from kirin import ir, types
from kirin.prelude import basic
from kirin.dialects import py, func, module


def test_module():
    fn1 = func.Function(
        sym_name="foo",
        slots=(),
        body=ir.Region(
            ir.Block(
                [
                    x := py.Constant(1),
                    func.Return(x.result),
                ],
                argtypes=(types.Any,),
            )
        ),
        signature=func.Signature(inputs=(), output=types.Int),
    )

    fn2 = func.Function(
        sym_name="main",
        slots=(),
        body=ir.Region(
            ir.Block(
                [
                    x := module.Invoke((), (), callee="foo"),
                    func.Return(x.result),
                ],
                argtypes=(types.Any,),
            )
        ),
        signature=func.Signature(inputs=(), output=types.Int),
    )

    mod = module.Module(
        sym_name="test_module", entry="main", body=ir.Region(ir.Block([fn1, fn2]))
    )

    dialects = basic.add(module)
    method = ir.Method(dialects=dialects, code=mod)
    method.print()

    with pytest.raises(KeyError):
        method()

    dialects.update_symbol_table(method)
    assert method() == 1



"""Definition of Kirin's Intermediate Representation (IR) nodes."""

from kirin.ir.nodes.base import IRNode as IRNode
from kirin.ir.nodes.stmt import Statement as Statement
from kirin.ir.nodes.block import Block as Block
from kirin.ir.nodes.region import Region as Region



from kirin import ir
from kirin.rewrite import abc
from kirin.analysis import const
from kirin.dialects import py

from ..stmts import New


class InlineGetItem(abc.RewriteRule):
    """Rewrite rule to inline GetItem statements for IList.

    For example if we have an `ilist.New` statement with a list of items,
    and we can infer that the index used in `py.GetItem` is constant and within bounds,
    we replace the `py.GetItem` with the ssa value in the list when the index is an integer
    or with a new `ilist.New` statement containing the sliced items when the index is a slice.

    """

    def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
        if not isinstance(node, py.GetItem) or not isinstance(
            stmt := node.obj.owner, New
        ):
            return abc.RewriteResult()

        if not isinstance(index_const := node.index.hints.get("const"), const.Value):
            return abc.RewriteResult()

        index = index_const.data
        if isinstance(index, int) and (
            0 <= index < len(stmt.args) or -len(stmt.args) <= index < 0
        ):
            node.result.replace_by(stmt.args[index])
            return abc.RewriteResult(has_done_something=True)
        elif isinstance(index, slice):
            start, stop, step = index.indices(len(stmt.args))
            new_tuple = New(
                tuple(stmt.args[start:stop:step]),
            )
            node.replace_by(new_tuple)
            return abc.RewriteResult(has_done_something=True)
        else:
            return abc.RewriteResult()



from kirin import ir, types
from kirin.interp import MethodTable, ReturnValue, StatementResult, impl
from kirin.analysis import const
from kirin.dialects.func.stmts import Call, Invoke, Lambda, Return, GetField

from ._dialect import dialect


@dialect.register(key="constprop")
class DialectConstProp(MethodTable):

    @impl(Return)
    def return_(
        self, interp: const.Propagate, frame: const.Frame, stmt: Return
    ) -> StatementResult[const.Result]:
        return ReturnValue(frame.get(stmt.value))

    @impl(Call)
    def call(
        self, interp: const.Propagate, frame: const.Frame, stmt: Call
    ) -> StatementResult[const.Result]:
        # give up on dynamic method calls
        callee = frame.get(stmt.callee)
        if isinstance(callee, const.PartialLambda):
            call_frame, ret = interp.call(
                callee.code,
                callee,
                *frame.get_values(stmt.inputs),
                **{k: v for k, v in zip(stmt.keys, frame.get_values(stmt.kwargs))},
            )
            if not call_frame.frame_is_not_pure:
                frame.should_be_pure.add(stmt)
            return (ret,)

        if not (isinstance(callee, const.Value) and isinstance(callee.data, ir.Method)):
            return (const.Result.bottom(),)

        mt: ir.Method = callee.data
        call_frame, ret = interp.call(
            mt.code,
            interp.method_self(mt),
            *frame.get_values(stmt.inputs),
            **{k: v for k, v in zip(stmt.keys, frame.get_values(stmt.kwargs))},
        )
        if not call_frame.frame_is_not_pure:
            frame.should_be_pure.add(stmt)
        return (ret,)

    @impl(Invoke)
    def invoke(
        self,
        interp: const.Propagate,
        frame: const.Frame,
        stmt: Invoke,
    ) -> StatementResult[const.Result]:
        invoke_frame, ret = interp.call(
            stmt.callee.code,
            interp.method_self(stmt.callee),
            *frame.get_values(stmt.inputs),
        )
        if not invoke_frame.frame_is_not_pure:
            frame.should_be_pure.add(stmt)
        return (ret,)

    @impl(Lambda)
    def lambda_(
        self, interp: const.Propagate, frame: const.Frame, stmt: Lambda
    ) -> StatementResult[const.Result]:
        captured = frame.get_values(stmt.captured)
        if stmt.body.blocks and types.is_tuple_of(captured, const.Value):
            return (
                const.Value(
                    ir.Method(
                        dialects=interp.dialects,
                        code=stmt,
                        fields=tuple(each.data for each in captured),
                    )
                ),
            )

        return (
            const.PartialLambda(
                stmt,
                tuple(each for each in captured),
            ),
        )

    @impl(GetField)
    def getfield(
        self,
        interp: const.Propagate,
        frame: const.Frame,
        stmt: GetField,
    ) -> StatementResult[const.Result]:
        callee_self = frame.get(stmt.obj)
        if isinstance(callee_self, const.Value) and isinstance(
            callee_self.data, ir.Method
        ):
            mt: ir.Method = callee_self.data
            return (const.Value(mt.fields[stmt.field]),)
        elif isinstance(callee_self, const.PartialLambda):
            return (callee_self.captured[stmt.field],)
        return (const.Unknown(),)



from __future__ import annotations

from kirin import emit, interp

from .stmts import Range
from ._dialect import dialect


@dialect.register(key="emit.julia")
class JuliaMethodTable(interp.MethodTable):

    @interp.impl(Range)
    def range(self, emit_: emit.Julia, frame: emit.JuliaFrame, node: Range):
        start = frame.get(node.start)
        stop = frame.get(node.stop)
        step = frame.get(node.step)
        frame.write_line(f"{frame.ssa[node.result]} = {start}:{step}:{stop}")
        return (frame.ssa[node.result],)



from kirin import ir
from kirin.interp import MethodTable, impl
from kirin.analysis import const

from .stmts import Map, Scan, Foldl, Foldr, ForEach
from ._dialect import dialect


@dialect.register(key="constprop")
class ConstPropMethods(MethodTable):

    def detect_purity(
        self,
        constprop: const.Propagate,
        frame: const.Frame,
        stmt: ir.Statement,
        method_code: ir.Statement,
        args: tuple[const.Result, ...],
    ):
        method_frame, _ = constprop.call(method_code, *args)
        if not method_frame.frame_is_not_pure:
            frame.should_be_pure.add(stmt)

    @impl(Map)
    @impl(ForEach)
    def one_args(
        self, interp_: const.Propagate, frame: const.Frame, stmt: ForEach | Map
    ):
        fn, collection = frame.get(stmt.fn), frame.get(stmt.collection)

        # 1. if the function is a constant method, and the method is pure, then the map is pure
        if isinstance(fn, const.Value) and isinstance(method := fn.data, ir.Method):
            self.detect_purity(interp_, frame, stmt, method.code, (fn, const.Unknown()))
            if isinstance(collection, const.Value):
                return interp_.try_eval_const_pure(frame, stmt, (fn, collection))
        elif isinstance(fn, const.PartialLambda):
            self.detect_purity(interp_, frame, stmt, fn.code, (fn, const.Unknown()))

        return (const.Unknown(),)

    @impl(Foldl)
    @impl(Foldr)
    @impl(Scan)
    def two_args(self, interp_: const.Propagate, frame: const.Frame, stmt: Foldl):
        fn, collection, init = (
            frame.get(stmt.fn),
            frame.get(stmt.collection),
            frame.get(stmt.init),
        )

        # 1. if the function is a constant method, and the method is pure, then the foldl is pure
        if isinstance(fn, const.Value) and isinstance(method := fn.data, ir.Method):
            self.detect_purity(
                interp_,
                frame,
                stmt,
                method.code,
                (fn, const.Unknown(), const.Unknown()),
            )
            if isinstance(collection, const.Value) and isinstance(init, const.Value):
                return interp_.try_eval_const_pure(frame, stmt, (fn, collection, init))
        elif isinstance(fn, const.PartialLambda):
            self.detect_purity(
                interp_, frame, stmt, fn.code, (fn, const.Unknown(), const.Unknown())
            )

        return (const.Unknown(),)



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Generic, TypeVar, Callable, TypeAlias, overload
from dataclasses import dataclass

from kirin import ir, types

if TYPE_CHECKING:
    from .abc import InterpreterABC
    from .frame import FrameABC
    from .value import RegionResult, StatementResult

MethodTableType = TypeVar("MethodTableType", bound="MethodTable")
InterpreterType = TypeVar("InterpreterType", bound="InterpreterABC")
FrameType = TypeVar("FrameType", bound="FrameABC")
ValueType = TypeVar("ValueType")
Head = TypeVar("Head")
Ret = TypeVar("Ret")
ClassMethod: TypeAlias = Callable[
    [
        MethodTableType,
        InterpreterType,
        FrameType,
        Head,
    ],
    Ret,
]


@dataclass(frozen=True)
class Signature:
    head: type | ir.RegionInterpretationTrait
    args: tuple = ()


NodeType = TypeVar("NodeType")


@dataclass
class Def(Generic[MethodTableType, InterpreterType, FrameType, NodeType, Ret]):
    signature: tuple[Signature, ...]
    method: ClassMethod[MethodTableType, InterpreterType, FrameType, NodeType, Ret]

    def __init__(
        self,
        signature: Signature,
        func: (
            Def[MethodTableType, InterpreterType, FrameType, NodeType, Ret]
            | ClassMethod[MethodTableType, InterpreterType, FrameType, NodeType, Ret]
        ),
    ):
        if isinstance(func, Def):
            assert (
                signature.args is not None
            ), f"Signature with no parameters {signature} cannot have multiple implementations"
            self.signature = func.signature + (signature,)
            self.method = func.method
        else:
            self.signature = (signature,)
            self.method = func

        self.__name__ = func.__name__
        self.__doc__ = func.__doc__

    def __get__(
        self,
        method_table: MethodTableType | None,
        owner: type[MethodTableType] | None = None,
    ):
        if method_table is None:
            raise TypeError(
                f"cannot use impl decorator on {self.method}, only on MethodTable"
            )

        if owner is None:
            owner = type(method_table)

        if not issubclass(owner, MethodTable):
            raise TypeError(
                f"cannot use impl decorator on {owner}, only on MethodTable"
            )
        return BoundedDef(method_table, self.signature, self.method)


@dataclass
class BoundedDef(Generic[MethodTableType, InterpreterType, FrameType, NodeType, Ret]):
    parent: MethodTableType
    signature: tuple[Signature, ...]
    method: ClassMethod[MethodTableType, InterpreterType, FrameType, NodeType, Ret]

    def __call__(
        self,
        interpreter: InterpreterType,
        frame: FrameType,
        node: NodeType,
    ) -> Ret:
        return self.method(self.parent, interpreter, frame, node)

    def __repr__(self) -> str:
        return f"impl {self.method.__name__} in {repr(self.parent.__class__)}"


@overload
def impl(head: ir.RegionInterpretationTrait) -> Callable[
    [
        ClassMethod[
            MethodTableType,
            InterpreterType,
            FrameType,
            ir.Region,
            RegionResult[ValueType],
        ]
        | Def[
            MethodTableType,
            InterpreterType,
            FrameType,
            ir.Region,
            RegionResult[ValueType],
        ]
    ],
    Def[
        MethodTableType, InterpreterType, FrameType, ir.Region, RegionResult[ValueType]
    ],
]: ...


AttributeType = TypeVar("AttributeType", bound=ir.Attribute)


@overload
def impl(head: type[AttributeType]) -> Callable[
    [
        ClassMethod[
            MethodTableType,
            InterpreterType,
            FrameType,
            AttributeType,
            ValueType,
        ]
        | Def[
            MethodTableType,
            InterpreterType,
            FrameType,
            AttributeType,
            ValueType,
        ]
    ],
    Def[
        MethodTableType,
        InterpreterType,
        FrameType,
        AttributeType,
        ValueType,
    ],
]: ...


StatementType = TypeVar("StatementType", bound=ir.Statement)
MethodStatementType = TypeVar("MethodStatementType", bound=ir.Statement)


@overload
def impl(head: type[StatementType], *params: types.TypeAttribute) -> Callable[
    [
        ClassMethod[
            MethodTableType,
            InterpreterType,
            FrameType,
            MethodStatementType,
            StatementResult[ValueType],
        ]
        | Def[
            MethodTableType,
            InterpreterType,
            FrameType,
            MethodStatementType,
            StatementResult[ValueType],
        ]
    ],
    Def[
        MethodTableType,
        InterpreterType,
        FrameType,
        MethodStatementType,
        StatementResult[ValueType],
    ],
]: ...


def impl(head, *params):
    def wrapper(func: ClassMethod, /):
        return Def(Signature(head, params), func)

    if isinstance(head, ir.RegionInterpretationTrait) or issubclass(
        head, (ir.Statement, ir.Attribute)
    ):
        return wrapper
    else:
        raise TypeError(f"Unsupported type: {head}")


class MethodTable(ABC):
    """Base class to define lookup tables for interpreting code for IR nodes in a dialect."""

    pass



from typing import Any

from kirin import types
from kirin.prelude import basic
from kirin.dialects.ilist import IList, IListType


@basic(typeinfer=True)
def tuple_vararg(xs: tuple[int, ...], i: int):
    return xs[i]


@basic(typeinfer=True)
def tuple_multi(xs: tuple[int, float, str], i: int):
    return xs[i]


@basic(typeinfer=True)
def tuple_slice(xs: tuple[int, float, str], i: slice):
    return xs[i]


@basic(typeinfer=True)
def tuple_vararg_slice(xs: tuple[int, ...], i: slice):
    return xs[i]


@basic(typeinfer=True)
def tuple_const_slice(xs: tuple[int, float, str]):
    return xs[1:]


@basic(typeinfer=True)
def tuple_const_index(xs: tuple[int, float, str]):
    return xs[1]


@basic(typeinfer=True)
def tuple_err(xs: tuple[int, float, str], i: str):
    return xs[i]


@basic(typeinfer=True)
def tuple_const_err(xs: tuple[int, float, str]):
    return xs[3]


@basic(typeinfer=True)
def list_infer(xs: IList[int, Any], i: int):
    return xs[i]


@basic(typeinfer=True)
def list_slice(xs: IList[int, Any], i: slice):
    return xs[i]


@basic(typeinfer=True)
def unknown(xs, i: int):
    return xs[i]


def test_getitem_typeinfer():
    assert tuple_vararg.return_type.is_subseteq(types.Int)
    assert tuple_multi.return_type.is_subseteq(types.Int | types.Float | types.String)
    assert tuple_slice.return_type.is_subseteq(
        types.Tuple[types.Vararg(types.Int | types.Float | types.String)]
    )
    assert tuple_const_index.return_type.is_subseteq(types.Float)
    assert tuple_vararg_slice.return_type.is_subseteq(
        types.Tuple[types.Vararg(types.Int)]
    )
    assert tuple_const_slice.return_type.is_subseteq(
        types.Tuple[types.Float, types.String]
    )
    assert tuple_err.return_type.is_structurally_equal(types.Bottom)
    assert tuple_const_err.return_type.is_structurally_equal(types.Bottom)
    assert list_infer.return_type.is_subseteq(types.Int)
    assert list_slice.return_type.is_subseteq(IListType[types.Int])
    assert unknown.return_type.is_structurally_equal(types.Any)



from typing import Any, cast

from kirin import ir, types
from kirin.dialects.func.attrs import Signature
from kirin.dialects.ilist.runtime import IList
from kirin.serialization.base.context import (
    MethodSymbolMeta,
    SerializationContext,
    mangle,
    get_cls_from_name,
)
from kirin.serialization.base.deserializable import Deserializable
from kirin.serialization.base.serializationunit import SerializationUnit
from kirin.serialization.base.serializationmodule import SerializationModule


class Deserializer:
    _ctx: SerializationContext

    def __init__(self) -> None:
        self._ctx = SerializationContext()
        self._ctx.clear()

    def decode(self, data: SerializationModule) -> ir.Method:
        for mangled, meta in data.symbol_table.items():
            sym_name = meta.get("sym_name", None)
            if sym_name is None:
                raise ValueError(f"symbol_table[{mangled}] missing 'sym_name'")
            arg_types = meta.get("arg_types", []) or []
            self._ctx.Method_Symbol[mangled] = MethodSymbolMeta(
                sym_name=sym_name,
                arg_types=list(arg_types),
            )

        body = data.body
        if body is None:
            raise ValueError("Module envelope missing body for decoding.")
        return self.deserialize_method(body)

    def deserialize(self, serUnit: SerializationUnit) -> Any:
        ser_method = getattr(
            self, "serialize_" + serUnit.class_name.lower(), self.generic_deserialize
        )
        return ser_method(serUnit)

    def generic_deserialize(self, data: SerializationUnit) -> Any:
        if not hasattr(data, "kind"):
            raise ValueError(
                f"Invalid SerializationUnit: {data} missing 'kind' attribute."
            )
        match data.kind:
            case "bool":
                return self.deserialize_boolean(data)
            case "bytes":
                return self.deserialize_bytes(data)
            case "bytearray":
                return self.deserialize_bytearray(data)
            case "dict":
                return self.deserialize_dict(data)
            case "float":
                return self.deserialize_float(data)
            case "frozenset":
                return self.deserialize_frozenset(data)
            case "int":
                return self.deserialize_int(data)
            case "list":
                return self.deserialize_list(data)
            case "range":
                return self.deserialize_range(data)
            case "set":
                return self.deserialize_set(data)
            case "slice":
                return self.deserialize_slice(data)
            case "str":
                return self.deserialize_str(data)
            case "none":
                return self.deserialize_none(data)
            case "tuple":
                return self.deserialize_tuple(data)
            case "type":
                return self.deserialize_type(data)
            case _:
                obj = get_cls_from_name(serUnit=data)
                if isinstance(obj, Deserializable):
                    return obj.deserialize(data, self)
                else:
                    raise ValueError(
                        f"Unsupported kind {data.kind} for deserialization."
                    )

    def deserialize_method(self, serUnit: SerializationUnit) -> ir.Method:
        mangled = serUnit.data.get("mangled")
        if mangled is None:
            raise ValueError("Missing 'mangled' key for method deserialization.")

        out = self._ctx.Method_Runtime.get(mangled)
        if out is None:
            out = ir.Method.__new__(ir.Method)
            out.mod = None
            out.py_func = None
            out.code = ir.Statement.__new__(ir.Statement)
            self._ctx.Method_Runtime[mangled] = out

        out.sym_name = serUnit.data["sym_name"]
        out.arg_names = serUnit.data.get("arg_names", [])
        out.dialects = self.deserialize_dialect_group(serUnit.data["dialects"])
        out.code = self.deserialize_statement(serUnit.data["code"])
        computed = mangle(
            out.sym_name,
            getattr(out, "arg_types", ()),
            getattr(out, "ret_type", None),
        )
        if computed != mangled:
            raise ValueError(
                f"Mangled name mismatch: expected {mangled}, got {computed}"
            )
        return out

    def deserialize_statement(self, serUnit: SerializationUnit) -> ir.Statement:
        cls = get_cls_from_name(serUnit=serUnit)
        data = serUnit.data
        out = ir.Statement.__new__(cls)
        self._ctx.Statement_Lookup[data["id"]] = out
        out.dialect = self.deserialize(data["dialect"])
        out.name = self.deserialize_str(data["name"])
        out._args = self.deserialize_tuple(data["_args"])
        out._results = self.deserialize_list(data["_results"])
        out._name_args_slice = self.deserialize_dict(data["_name_args_slice"])
        out.attributes = self.deserialize_dict(data["attributes"])
        out.successors = self.deserialize_list(data["successors"])
        _regions = self.deserialize_list(data["_regions"])
        for region in _regions:
            if region.parent_node is None:
                region.parent_node = out
        out._regions = _regions
        return out

    def deserialize_blockargument(self, serUnit: SerializationUnit) -> ir.BlockArgument:
        cls = get_cls_from_name(serUnit=serUnit)
        ssa_name = serUnit.data["id"]
        if ssa_name in self._ctx.SSA_Lookup:
            existing = self._ctx.SSA_Lookup[ssa_name]
            if isinstance(existing, ir.BlockArgument):
                return existing
            raise ValueError(
                f"Block argument id {ssa_name} already present but maps to {type(existing).__name__}"
            )

        blk_name = serUnit.data["blk_id"]
        if blk_name not in self._ctx.Block_Lookup:
            block = ir.Block.__new__(ir.Block)
            self._ctx.Block_Lookup[blk_name] = block
        else:
            block = self._ctx.Block_Lookup[blk_name]
        index = serUnit.data["index"]
        typ = self.deserialize_attribute(serUnit.data["type"])
        if not isinstance(typ, types.TypeAttribute):
            raise TypeError(f"Expected a TypeAttribute, got {type(typ)!r}: {typ!r}")
        out = cls(block=block, index=index, type=cast(types.TypeAttribute, typ))
        out._name = serUnit.data.get("name", None)
        self._ctx.SSA_Lookup[ssa_name] = out

        return out

    def deserialize_region(self, serUnit: SerializationUnit) -> ir.Region:
        if serUnit.kind == "region":
            out = ir.Region.__new__(ir.Region)
            region_name = serUnit.data.get("id")
            if region_name is not None:
                self._ctx.Region_Lookup[region_name] = out

            blocks = [self.deserialize(blk) for blk in serUnit.data.get("blocks", [])]

            out._blocks = []
            out._block_idx = {}

            for block in blocks:
                existing_parent = block.parent
                if existing_parent is not None and existing_parent is not out:
                    block.parent = None
                out.blocks.append(block)

            return out
        elif serUnit.data.get("kind") == "region_ref":
            region_name = serUnit.data["id"]
            if region_name not in self._ctx.Region_Lookup:
                raise ValueError(f"Region with id {region_name} not found in lookup.")
            return self._ctx.Region_Lookup[region_name]
        else:
            raise ValueError("Invalid region data for decoding.")

    def deserialize_block(self, serUnit: SerializationUnit) -> ir.Block:
        if serUnit.kind == "block_ref":
            return self.deserialize_block_ref(serUnit)
        elif serUnit.kind == "block":
            return self.deserialize_concrete_block(serUnit)
        else:
            raise ValueError("Invalid block data for decoding.")

    def deserialize_block_ref(self, serUnit: SerializationUnit) -> ir.Block:
        if serUnit.kind != "block_ref":
            raise ValueError("Invalid block reference data for decoding.")

        block_name = serUnit.data["id"]
        if block_name not in self._ctx.Block_Lookup:
            raise ValueError(f"Block with id {block_name} not found in lookup.")
        return self._ctx.Block_Lookup[block_name]

    def deserialize_concrete_block(self, serUnit: SerializationUnit) -> ir.Block:
        if serUnit.kind != "block":
            raise ValueError("Invalid block data for decoding.")

        block_name = serUnit.data["id"]

        if block_name not in self._ctx.Block_Lookup:
            if block_name in self._ctx._block_reference_store:
                out = self._ctx._block_reference_store.pop(block_name)
                self._ctx.Block_Lookup[block_name] = out
            else:
                out = ir.Block.__new__(ir.Block)
                self._ctx.Block_Lookup[block_name] = out
        else:
            out = self._ctx.Block_Lookup[block_name]

        out._args = tuple(
            self.deserialize_blockargument(arg_data)
            for arg_data in serUnit.data.get("_args", [])
        )

        stmts_data = serUnit.data.get("stmts")
        if stmts_data is None:
            raise ValueError("Block data must contain 'stmts' field.")

        out._first_stmt = None
        out._last_stmt = None
        out._first_branch = None
        out._last_branch = None
        out._stmt_len = 0
        stmts = tuple(self.deserialize_statement(stmt_data) for stmt_data in stmts_data)
        out.stmts.extend(stmts)

        return out

    def deserialize_boolean(self, serUnit: SerializationUnit) -> bool:
        return bool(serUnit.data["value"])

    def deserialize_bytes(self, serUnit: SerializationUnit) -> bytes:
        return bytes.fromhex(serUnit.data["value"])

    def deserialize_bytearray(self, serUnit: SerializationUnit) -> bytearray:
        return bytearray.fromhex(serUnit.data["value"])

    def deserialize_dict(self, serUnit: SerializationUnit) -> dict:
        keys = [self.deserialize(k) for k in serUnit.data.get("keys", [])]
        values = [self.deserialize(v) for v in serUnit.data.get("values", [])]
        return dict(zip(keys, values))

    def deserialize_float(self, serUnit: SerializationUnit) -> float:
        return float(serUnit.data["value"])

    def deserialize_frozenset(self, serUnit: SerializationUnit) -> frozenset:
        return frozenset(self.deserialize(x) for x in serUnit.data.get("value", []))

    def deserialize_int(self, serUnit: SerializationUnit) -> int:
        return int(serUnit.data["value"])

    def deserialize_list(self, serUnit: SerializationUnit) -> list:
        return [self.deserialize(x) for x in serUnit.data.get("value", [])]

    def deserialize_range(self, serUnit: SerializationUnit) -> range:
        start = self.deserialize(serUnit.data.get("start", 0))
        stop = self.deserialize(serUnit.data.get("stop", 0))
        step = self.deserialize(serUnit.data.get("step", 1))
        return range(start, stop, step)

    def deserialize_set(self, serUnit: SerializationUnit) -> set:
        return set(self.deserialize(x) for x in serUnit.data.get("value", []))

    def deserialize_slice(self, serUnit: SerializationUnit) -> slice:
        start = self.deserialize(serUnit.data["start"])
        stop = self.deserialize(serUnit.data["stop"])
        step = self.deserialize(serUnit.data["step"])
        return slice(start, stop, step)

    def deserialize_str(self, serUnit: SerializationUnit) -> str:
        return serUnit.data["value"]

    def deserialize_none(self, serUnit: SerializationUnit) -> None:
        return None

    def deserialize_tuple(self, serUnit: SerializationUnit) -> tuple:
        return tuple(self.deserialize(x) for x in serUnit.data.get("value", []))

    def deserialize_attribute(self, serUnit: SerializationUnit) -> ir.Attribute:
        cls = get_cls_from_name(serUnit)
        if not issubclass(cls, Deserializable):
            raise TypeError(f"Class {cls} is not Deserializable.")
        assert issubclass(cls, ir.Attribute)
        return cls.deserialize(serUnit, self)

    def deserialize_resultvalue(self, serUnit: SerializationUnit) -> ir.ResultValue:
        ssa_name = serUnit.data["id"]
        if ssa_name in self._ctx.SSA_Lookup:
            existing = self._ctx.SSA_Lookup[ssa_name]
            if isinstance(existing, ir.ResultValue):
                return existing
            raise ValueError(
                f"SSA id {ssa_name} already exists and is {type(existing).__name__}"
            )
        index = int(serUnit.data["index"])

        typ = self.deserialize_attribute(serUnit.data["type"])
        if typ is None or not isinstance(typ, types.TypeAttribute):
            raise TypeError(f"Expected a TypeAttribute, got {type(typ)!r}: {typ!r}")
        owner: ir.Statement = self._ctx.Statement_Lookup[
            self.deserialize_str(serUnit.data["owner"])
        ]
        out = ir.ResultValue(
            stmt=owner, index=index, type=cast(types.TypeAttribute, typ)
        )
        out.name = serUnit.data.get("name", None)

        self._ctx.SSA_Lookup[ssa_name] = out

        return out

    def deserialize_type(self, serUnit: SerializationUnit) -> type:
        cls = get_cls_from_name(serUnit)
        return cls

    def deserialize_dialect(self, serUnit: SerializationUnit) -> ir.Dialect:
        name = self.deserialize_str(serUnit.data["name"])
        stmts = self.deserialize_list(serUnit.data["stmts"])
        cls = get_cls_from_name(serUnit)
        return cls(name=name, stmts=stmts)

    def deserialize_dialect_group(self, serUnit: SerializationUnit) -> ir.DialectGroup:
        dialects = self.deserialize_frozenset(serUnit.data["data"])
        cls = get_cls_from_name(serUnit)
        return cls(dialects=dialects)

    def deserialize_pyclass(self, serUnit: SerializationUnit) -> types.PyClass:
        return types.PyClass(
            typ=self.deserialize(serUnit.data["typ"]),
            display_name=self.deserialize(serUnit.data.get("display_name", "")),
            prefix=self.deserialize(serUnit.data.get("prefix", "")),
        )

    def deserialize_typevar(self, serUnit: SerializationUnit) -> types.TypeVar:
        varname = self.deserialize(serUnit.data["varname"])
        bound = self.deserialize(serUnit.data["bound"])
        return types.TypeVar(varname, bound)

    def deserialize_generic(self, serUnit: SerializationUnit) -> types.Generic:
        body = self.deserialize_pyclass(serUnit.data["body"])
        vars = self.deserialize_tuple(serUnit.data["vars"])
        vararg = self.deserialize(serUnit.data["vararg"])
        out = types.Generic(body, *vars)
        out.vararg = vararg
        return out

    def deserialize_vararg(self, serUnit: SerializationUnit) -> types.Vararg:
        typ = self.deserialize(serUnit.data["typ"])
        return types.Vararg(typ)

    def deserialize_anytype(self, serUnit: SerializationUnit) -> types.AnyType:
        return types.AnyType()

    def deserialize_bottomtype(self, serUnit: SerializationUnit) -> types.BottomType:
        return types.BottomType()

    def deserialize_pyattr(self, serUnit: SerializationUnit) -> ir.PyAttr:
        pytype = self.deserialize(serUnit.data["pytype"])
        value = self.deserialize(serUnit.data["data"])
        return ir.PyAttr(value, pytype=pytype)

    def deserialize_literal(self, serUnit: SerializationUnit) -> types.Literal:
        d = self.deserialize(serUnit.data["value"])
        type_attr = self.deserialize(serUnit.data["type"])
        return types.Literal(d, type_attr)

    def deserialize_union(self, serUnit: SerializationUnit) -> types.Union:
        ty = self.deserialize_frozenset(serUnit.data["types"])
        return types.Union(ty)

    def deserialize_signature(self, serUnit: SerializationUnit) -> Signature:
        inputs = self.deserialize(serUnit.data["inputs"])
        output = self.deserialize(serUnit.data["output"])
        return Signature(inputs=inputs, output=output)

    def deserialize_ilist(self, serUnit: SerializationUnit) -> IList:
        items = self.deserialize(serUnit.data["data"])
        elem = self.deserialize(serUnit.data["elem"])
        return IList(items, elem=elem)



from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.passes import Fold, TypeInfer
from kirin.prelude import structural_no_opt
from kirin.analysis import const
from kirin.dialects import py, func, ilist

dialect = ir.Dialect("analysis")


@statement(dialect=dialect)
class MultiReturnStatement(ir.Statement):
    traits = frozenset({lowering.FromPythonCall(), ir.Pure()})
    inputs: tuple[ir.SSAValue] = info.argument(types.Any)

    def __init__(self, *args: ir.SSAValue):
        super().__init__(
            args=args,
            result_types=tuple(arg.type for arg in args),
            args_slice={"inputs": slice(None)},
        )


@statement(dialect=dialect)
class MultiReturnNotPure(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    inputs: tuple[ir.SSAValue] = info.argument(types.Any)

    def __init__(self, *args: ir.SSAValue):
        super().__init__(
            args=args,
            result_types=tuple(arg.type for arg in args),
            args_slice={"inputs": slice(None)},
        )


@ir.dialect_group(structural_no_opt.add(dialect))
def dialect_group_test(self):
    fold = Fold(self)
    type_infer = TypeInfer(self)

    def run_pass(mt):
        type_infer(mt)
        fold(mt)

    return run_pass


def test_multi_return_default_prop():

    stmts = [
        (a := py.Constant(1)),
        (b := py.Constant(2)),
        (res := MultiReturnStatement(a.result, b.result)),
        (return_result := ilist.New((res.results[0], res.results[1]))),
        (res := MultiReturnNotPure(a.result, b.result)),
        (func.Return(return_result.result)),
    ]

    body = ir.Region(ir.Block(stmts, argtypes=(types.Any,)))
    func_code = func.Function(
        sym_name="test", signature=func.Signature((), types.Any), body=body
    )
    mt = ir.Method(dialects=dialect_group_test, code=func_code)
    frame, return_result = const.Propagate(dialect_group_test).run(mt)

    assert frame.entries[res.results[0]] == const.Unknown()
    assert frame.entries[res.results[1]] == const.Unknown()



from kirin.prelude import structural
from kirin.dialects import py, func
from kirin.passes.aggressive import UnrollScf


def test_unroll_scf():
    @structural
    def main(r: list[int], cond: bool):
        if cond:
            for i in range(4):
                tmp = r[-1]
                if i < 2:
                    tmp += i * 2
                else:
                    for j in range(4):
                        if i > j:
                            tmp += i + j
                        else:
                            tmp += i - j

                r.append(tmp)
        else:
            for i in range(4):
                r.append(i)
        return r

    UnrollScf(structural).fixpoint(main)

    num_adds = 0
    num_calls = 0

    for op in main.callable_region.walk():
        if isinstance(op, py.Add):
            num_adds += 1
        elif isinstance(op, func.Call):
            num_calls += 1

    assert num_adds == 10
    assert num_calls == 8



# type: ignore
from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.passes import Fold
from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk, Fixpoint, WrapConst
from kirin.analysis import const
from kirin.dialects.py import constant
from kirin.rewrite.dce import DeadCodeElimination
from kirin.rewrite.fold import ConstantFold
from kirin.rewrite.inline import Inline
from kirin.rewrite.getfield import InlineGetField
from kirin.rewrite.compactify import CFGCompactify

fold_pass = Fold(basic_no_opt)


@basic_no_opt
def somefunc(x: int):
    return x - 1


@basic_no_opt
def main(x: int):
    return somefunc(x) + 1


def test_simple():
    inline = Inline(heuristic=lambda x: True)
    a = main(1)
    main.code.print()
    Walk(inline).rewrite(main.code)
    main.code.print()
    b = main(1)
    assert a == b


@basic_no_opt
def closure_double(x: int, y: int):
    def foo(a: int, b: int):
        return a + b + x + y

    return foo


@basic_no_opt
def inline_closure():
    a = 3
    b = 4
    c = closure_double(1, 2)
    return c(a, b) * 4


def test_inline_closure():
    fold_pass.fixpoint(inline_closure)
    Walk(Inline(heuristic=lambda x: True)).rewrite(inline_closure.code)
    fold_pass.fixpoint(inline_closure)
    inline_closure.code.print()
    stmt = inline_closure.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, constant.Constant)
    assert inline_closure() == 40


@basic_no_opt
def add(x, y):
    return x + y


@basic_no_opt
def foldl(f, acc, xs: tuple):
    if not xs:
        return acc
    ret = foldl(f, acc, xs[1:])
    return f(ret, xs[0])


@basic_no_opt
def inline_foldl(x):
    return foldl(add, 0, (x, x, x))


def test_inline_constprop():
    for _ in range(5):
        Walk(Inline(heuristic=lambda x: True)).rewrite(inline_foldl.code)
        fold_pass.fixpoint(inline_foldl)
    # inline_foldl.print(hint="const")
    assert len(inline_foldl.callable_region.blocks) == 1
    assert inline_foldl(2) == 6


def test_inline_single_entry():
    dialect = ir.Dialect("dummy2")

    @statement(dialect=dialect)
    class DummyStmtWithSiteEffect(ir.Statement):
        name = "dummy2"
        traits = frozenset({lowering.FromPythonCall()})
        value: ir.SSAValue = info.argument(types.Int)
        option: str = info.attribute()
        # result: ir.ResultValue = info.result(types.Int)

    @basic_no_opt.add(dialect)
    def inline_npure(x: int, y: int):
        DummyStmtWithSiteEffect(x, option="attr")
        DummyStmtWithSiteEffect(y, option="attr2")

    @basic_no_opt.add(dialect)
    def inline_non_pure():
        DummyStmtWithSiteEffect(3, option="attr0")
        inline_npure(1, 2)

    inline_non_pure.code.print()
    inline = Inline(heuristic=lambda x: True)
    Walk(inline).rewrite(inline_non_pure.code)
    Fixpoint(CFGCompactify()).rewrite(inline_non_pure.code)
    inline_non_pure.code.print()
    assert isinstance(
        inline_non_pure.callable_region.blocks[0].stmts.at(1), DummyStmtWithSiteEffect
    )
    assert isinstance(
        inline_non_pure.callable_region.blocks[0].stmts.at(5), DummyStmtWithSiteEffect
    )
    assert isinstance(
        inline_non_pure.callable_region.blocks[0].stmts.at(6), DummyStmtWithSiteEffect
    )


def test_inline_non_foldable_closure():
    dialect = ir.Dialect("dummy2")

    @statement(dialect=dialect)
    class DummyStmt2(ir.Statement):
        name = "dummy2"
        traits = frozenset({lowering.FromPythonCall()})
        value: ir.SSAValue = info.argument(types.Int)
        option: str = info.attribute()
        result: ir.ResultValue = info.result(types.Int)

    @basic_no_opt.add(dialect)
    def unfolable(x: int, y: int):
        def inner():
            DummyStmt2(x, option="hello")
            DummyStmt2(y, option="hello")

        return inner

    @basic_no_opt.add(dialect)
    def main():
        x = DummyStmt2(1, option="hello")
        x = unfolable(x, x)
        return x()

    main.print()
    inline = Walk(Inline(lambda _: True))
    inline.rewrite(main.code)
    constprop = const.Propagate(basic_no_opt)
    frame, _ = constprop.run(main)
    Walk(Fixpoint(WrapConst(frame))).rewrite(main.code)
    ConstantFold().rewrite(main.code)
    compact = Fixpoint(CFGCompactify())
    compact.rewrite(main.code)
    inline.rewrite(main.code)
    compact = Fixpoint(CFGCompactify())
    compact.rewrite(main.code)
    Fixpoint(Walk(InlineGetField())).rewrite(main.code)
    constprop = const.Propagate(basic_no_opt)
    frame, ret = constprop.run(main)
    Walk(Fixpoint(WrapConst(frame))).rewrite(main.code)
    Walk(DeadCodeElimination()).rewrite(main.code)
    main.print(analysis=frame.entries)

    @basic_no_opt.add(dialect)
    def target():
        x = DummyStmt2(1, option="hello")
        DummyStmt2(x, option="hello")
        DummyStmt2(x, option="hello")
        return

    CFGCompactify().rewrite(target.code)
    assert target.callable_region.is_structurally_equal(main.callable_region)



import pytest

from kirin.testing import assert_statements_same
from kirin.dialects import py


def test_same_statements_pass():
    statement_1 = py.Constant(1)
    statement_2 = py.Constant(1)
    assert_statements_same(statement_1, statement_2)


def test_different_statements_fail():
    statement_1 = py.Constant(1)
    statement_2 = py.Mult(statement_1.result, statement_1.result)
    with pytest.raises(AssertionError):
        assert_statements_same(statement_1, statement_2)


def test_same_statement_different_args_fails_1():
    statement_1 = py.Constant(1)
    statement_2 = py.Constant(2)
    with pytest.raises(AssertionError):
        assert_statements_same(statement_1, statement_2)


def test_same_statement_different_args_fails_2():
    arg_1 = py.Constant(1)
    arg_2 = py.Constant(2)
    statement_1 = py.Mult(arg_1.result, arg_1.result)
    statement_2 = py.Mult(arg_2.result, arg_2.result)
    with pytest.raises(AssertionError):
        assert_statements_same(statement_1, statement_2)


def test_same_statement_different_args_check_args_false_passes():
    arg_1 = py.Constant(1)
    arg_2 = py.Constant(2)
    statement_1 = py.Mult(arg_1.result, arg_1.result)
    statement_2 = py.Mult(arg_2.result, arg_2.result)
    assert_statements_same(statement_1, statement_2, check_args=False)



from kirin import lowering
from kirin.prelude import basic_no_opt
from kirin.dialects import cf, func


def test_issue_337():
    def test_if_inside_for() -> int:
        count = 0
        for i in range(5):
            count = count + 1
            if True:
                count = count + 100
            else:
                count = count + 300
        return count

    lower = lowering.Python(basic_no_opt)
    code = lower.python_function(test_if_inside_for, compactify=True)
    assert isinstance(code, func.Function)
    loop_last_block = code.body.blocks[4]
    count_5 = loop_last_block.args[0]
    stmt = loop_last_block.stmts.at(-1)
    assert isinstance(stmt, cf.ConditionalBranch)
    assert stmt.then_arguments[0] is count_5
    assert stmt.else_arguments[1] is count_5



import ast

from kirin import ir, types, lowering
from kirin.dialects import func

dialect = ir.Dialect("lowering.call")


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Call_local(
        self, state: lowering.State, callee: ir.SSAValue, node: ast.Call
    ) -> lowering.Result:
        source = state.source
        args, kwargs, keys = self.__lower_Call_args_kwargs(state, node)
        stmt = func.Call(callee, args, kwargs, keys=keys)
        stmt.source = source
        return state.current_frame.push(stmt)

    def lower_Call_global_method(
        self,
        state: lowering.State,
        method: ir.Method,
        node: ast.Call,
    ) -> lowering.Result:
        source = state.source
        args, kwargs, keys = self.__lower_Call_args_kwargs(state, node)
        inputs: list[ir.SSAValue] = [*args]
        kwargs_ = {k: v for k, v in zip(keys, kwargs)}

        if method.arg_names is None and keys:
            raise lowering.BuildError("method has no argument names, cannot use kwargs")

        if method.arg_names and keys:
            for name in method.arg_names:
                if name in keys:
                    inputs.append(kwargs_[name])

        stmt = func.Invoke(tuple(inputs), callee=method)
        stmt.result.type = method.return_type or types.Any
        stmt.source = source
        return state.current_frame.push(stmt)

    def __lower_Call_args_kwargs(
        self,
        state: lowering.State,
        node: ast.Call,
    ):
        args: list[ir.SSAValue] = []
        for arg in node.args:
            if isinstance(arg, ast.Starred):  # TODO: support *args
                raise lowering.BuildError("starred arguments are not supported")
            else:
                args.append(state.lower(arg).expect_one())

        keys: list[str] = []
        kwargs: list[ir.SSAValue] = []
        for kw in node.keywords:
            if kw.arg is None:
                raise lowering.BuildError("keyword argument must have a name")
            keys.append(kw.arg)
            kwargs.append(state.lower(kw.value).expect_one())

        return tuple(args), tuple(kwargs), tuple(keys)



import ast

from kirin import ir, types, lowering
from kirin.dialects import cf, func

dialect = ir.Dialect("lowering.func")


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Return(self, state: lowering.State, node: ast.Return) -> lowering.Result:
        if node.value is None:
            state.current_frame.push(
                func.Return(state.current_frame.push(func.ConstantNone()).result)
            )
        else:
            result = state.lower(node.value).expect_one()
            stmt = func.Return(result)
            state.current_frame.push(stmt)

    def lower_FunctionDef(
        self, state: lowering.State[ast.AST], node: ast.FunctionDef
    ) -> lowering.Result:
        self.assert_simple_arguments(node.args)
        signature = func.Signature(
            inputs=tuple(
                self.get_hint(state, arg.annotation) for arg in node.args.args
            ),
            output=self.get_hint(state, node.returns),
        )
        frame = state.current_frame

        slots = tuple(arg.arg for arg in node.args.args)
        entries: dict[str, ir.SSAValue] = {}
        entr_block = ir.Block()
        fn_self = entr_block.args.append_from(
            types.Generic(
                ir.Method, types.Tuple.where(signature.inputs), signature.output
            ),
            node.name + "_self",
        )
        entries[node.name] = fn_self
        for arg, type in zip(node.args.args, signature.inputs):
            entries[arg.arg] = entr_block.args.append_from(type, arg.arg)

        def callback(frame: lowering.Frame, value: ir.SSAValue):
            first_stmt = entr_block.first_stmt
            stmt = func.GetField(obj=fn_self, field=len(frame.captures) - 1)
            if value.name:
                stmt.result.name = value.name
            stmt.result.type = value.type
            stmt.source = state.source
            if first_stmt:
                stmt.insert_before(first_stmt)
            else:
                entr_block.stmts.append(stmt)
            return stmt.result

        with state.frame(
            node.body, entr_block=entr_block, capture_callback=callback
        ) as func_frame:
            func_frame.defs.update(entries)
            func_frame.exhaust()

            for block in func_frame.curr_region.blocks:
                if not block.last_stmt or not block.last_stmt.has_trait(
                    ir.IsTerminator
                ):
                    block.stmts.append(
                        cf.Branch(arguments=(), successor=func_frame.next_block)
                    )

            none_stmt = func.ConstantNone()
            rtrn_stmt = func.Return(none_stmt.result)
            func_frame.next_block.stmts.append(none_stmt)
            func_frame.next_block.stmts.append(rtrn_stmt)

        if state.current_frame.parent is None:  # toplevel function
            # TODO: allow a return value from function
            # and assign it to the function symbol
            frame.push(
                func.Function(
                    sym_name=node.name,
                    slots=slots,
                    signature=signature,
                    body=func_frame.curr_region,
                )
            )
            return

        if node.decorator_list:
            raise lowering.BuildError(
                "decorators are not supported on nested functions"
            )

        # nested function, lookup unknown variables
        first_stmt = func_frame.curr_region.blocks[0].first_stmt
        if first_stmt is None:
            raise lowering.BuildError("empty function body")

        lambda_stmt = func.Lambda(
            tuple(value for value in func_frame.captures.values()),
            sym_name=node.name,
            slots=slots,
            signature=signature,
            body=func_frame.curr_region,
        )
        lambda_stmt.result.name = node.name
        frame.push(lambda_stmt)
        # NOTE: Python automatically assigns the lambda to the name
        frame.defs[node.name] = lambda_stmt.result

    def assert_simple_arguments(self, node: ast.arguments) -> None:
        if node.kwonlyargs:
            raise lowering.BuildError("keyword-only arguments are not supported")

        if node.posonlyargs:
            raise lowering.BuildError("positional-only arguments are not supported")



import ast

import rich

from kirin import ir, decl, emit, types, interp, lowering

dialect = ir.Dialect("debug")


class InfoLowering(lowering.FromPythonCall):

    def lower(
        self, stmt: type, state: lowering.State, node: ast.Call
    ) -> lowering.Result:
        if len(node.args) == 0:
            raise lowering.BuildError("info() requires at least one argument")

        msg = state.lower(node.args[0]).expect_one()
        if len(node.args) > 1:
            inputs = tuple(state.lower(arg).expect_one() for arg in node.args[1:])
        else:
            inputs = ()
        state.current_frame.push(Info(msg=msg, inputs=inputs))


@decl.statement(dialect=dialect)
class Info(ir.Statement):
    """print debug information.

    This statement is used to print debug information during
    execution. The compiler has freedom to choose how to print
    the information and send it back to the caller. Note that
    in the case of heterogeneous hardware, this may not be printed
    on the same device as the caller but instead being a log.
    """

    traits = frozenset({InfoLowering()})
    msg: ir.SSAValue = decl.info.argument(types.String)
    inputs: tuple[ir.SSAValue, ...] = decl.info.argument()


@lowering.wraps(Info)
def info(msg: str, *inputs) -> None: ...


@dialect.register(key="main")
class ConcreteMethods(interp.MethodTable):

    @interp.impl(Info)
    def info(self, interp: interp.Interpreter, frame: interp.Frame, stmt: Info):
        # print("INFO:", frame.get(stmt.msg))
        rich.print(
            "[dim]┌───────────────────────────────────────────────────────────────[/dim]"
        )
        rich.print("[dim]│[/dim] [bold cyan]INFO:[/bold cyan] ", end="", sep="")
        print(frame.get(stmt.msg))
        for input in stmt.inputs:
            rich.print(
                "[dim]│[/dim] ",
                input.name or "unknown",
                "[dim] = [/dim]",
                end="",
                sep="",
            )
            print(frame.get(input))
        rich.print(
            "[dim]└───────────────────────────────────────────────────────────────[/dim]"
        )


@dialect.register(key="emit.julia")
class JuliaEmit(interp.MethodTable):
    @interp.impl(Info)
    def info(self, emit: emit.Julia, frame: emit.JuliaFrame, stmt: Info):
        msg = frame.get(stmt.msg)
        inputs = " ".join(frame.get(input) for input in stmt.inputs)
        frame.write_line(f'@info "{msg[1:-1]}" {inputs}'.strip())



from kirin import types, lowering
from kirin.dialects import cf, py, func
from kirin.dialects.lowering import func as func_lowering

lower = lowering.Python([cf, func, py.base, py.list, py.assign, func_lowering])


def test_empty_list():

    def empty_list():
        x = []
        return x

    code = lower.python_function(empty_list)

    list_stmt = code.body.blocks[0].stmts.at(0)  # type: ignore

    assert isinstance(list_stmt, py.list.New)
    assert len(list_stmt._results) == 1

    res = list_stmt._results[0]
    assert res.type.is_subseteq(types.List)



import typing

from kirin import types
from kirin.prelude import basic

N = typing.TypeVar("N")


class MyTest(typing.Generic[N]):
    pass


def test_generic_type_hint():
    @basic
    def test(obj: MyTest[N]):
        return None

    assert test.arg_types[0].is_subseteq(types.PyClass(MyTest))



from kirin.interp import Successor, MethodTable, impl
from kirin.analysis import const
from kirin.dialects.cf.stmts import Branch, ConditionalBranch
from kirin.dialects.cf.dialect import dialect


@dialect.register(key="constprop")
class ConstPropMethodTable(MethodTable):

    @impl(Branch)
    def branch(self, interp: const.Propagate, frame: const.Frame, stmt: Branch):
        interp.state.current_frame.worklist.append(
            Successor(stmt.successor, *frame.get_values(stmt.arguments))
        )
        return ()

    @impl(ConditionalBranch)
    def conditional_branch(
        self,
        interp: const.Propagate,
        frame: const.Frame,
        stmt: ConditionalBranch,
    ):
        frame = interp.state.current_frame
        cond = frame.get(stmt.cond)
        if isinstance(cond, const.Value):
            else_successor = Successor(
                stmt.else_successor, *frame.get_values(stmt.else_arguments)
            )
            then_successor = Successor(
                stmt.then_successor, *frame.get_values(stmt.then_arguments)
            )
            if cond.data:
                frame.worklist.append(then_successor)
            else:
                frame.worklist.append(else_successor)
        else:
            frame.entries[stmt.cond] = const.Value(True)
            then_successor = Successor(
                stmt.then_successor, *frame.get_values(stmt.then_arguments)
            )
            frame.worklist.append(then_successor)

            frame.entries[stmt.cond] = const.Value(False)
            else_successor = Successor(
                stmt.else_successor, *frame.get_values(stmt.else_arguments)
            )
            frame.worklist.append(else_successor)

            frame.entries[stmt.cond] = cond
        return ()



import pytest

from kirin import ir, types, lowering
from kirin.prelude import python_no_opt
from kirin.dialects import cf, func

lower = lowering.Python(python_no_opt)


def test_basic_func():
    def single(n):
        return n + 1

    code = lower.python_function(single)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    assert isinstance(code.body.blocks[0].last_stmt, func.Return)

    def single_2(n):
        return n + 1, n + 2

    code = lower.python_function(single_2)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    assert isinstance(code.body.blocks[0].last_stmt, func.Return)
    assert code.body.blocks[0].last_stmt.args[0].type.is_subseteq(types.Tuple)


def test_recursive_func():
    def recursive(n):
        if n == 0:
            return 0
        return recursive(n - 1)

    code = lower.python_function(recursive)
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 3
    assert isinstance(code.body.blocks[0].last_stmt, cf.ConditionalBranch)
    assert isinstance(code.body.blocks[2].stmts.at(2), func.Call)
    stmt: func.Call = code.body.blocks[2].stmts.at(2)  # type: ignore
    assert isinstance(stmt.callee, ir.BlockArgument)
    assert stmt.callee.type.is_subseteq(func.MethodType)


def test_invalid_func_call():

    def undefined(n):
        return foo(n - 1)  # type: ignore # noqa: F821

    with pytest.raises(lowering.BuildError):
        lower.python_function(undefined)

    def calling_python(n):
        return print(n)

    with pytest.raises(
        lowering.BuildError,
    ):
        lower.python_function(calling_python)


def test_func_call():
    def callee(n):
        return n + 1

    def caller(n):
        return callee(n)

    code = lower.python_function(callee)
    callee = ir.Method(
        dialects=lower.dialects, code=code, py_func=callee, sym_name="callee"
    )
    code = lower.python_function(caller, globals={"callee": callee})
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    stmt = code.body.blocks[0].stmts.at(0)
    assert isinstance(stmt, func.Invoke)
    assert isinstance(stmt.callee, ir.Method)


def test_func_kw_call():
    def callee(n, m):
        return n + m

    def caller(n, m):  # type: ignore
        return callee(n=n, m=m)

    code = lower.python_function(callee)
    callee = ir.Method(
        dialects=lower.dialects,
        code=code,
        py_func=callee,
        sym_name="callee",
        arg_names=["n", "m"],
    )
    code = lower.python_function(caller, globals={"callee": callee})
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    stmt = code.body.blocks[0].stmts.at(0)
    assert isinstance(stmt, func.Invoke)

    def caller(n, m):
        return callee(m=m, n=n)

    code = lower.python_function(caller, globals={"callee": callee})
    assert isinstance(code, func.Function)
    assert len(code.body.blocks) == 1
    stmt = code.body.blocks[0].stmts.at(0)
    assert isinstance(stmt, func.Invoke)
    assert len(stmt.inputs) == 2
    assert stmt.inputs[0] is code.body.blocks[0].args[1]
    assert stmt.inputs[1] is code.body.blocks[0].args[2]



from __future__ import annotations

from typing import TypeVar

from kirin import ir, interp, lattice

dialect = ir.Dialect("ssacfg")


@dialect.register(key="main")
class Concrete(interp.MethodTable):

    @interp.impl(ir.SSACFG())
    def ssacfg(self, interp_: interp.Interpreter, frame: interp.Frame, node: ir.Region):
        block = node.blocks[0]
        block_inputs = frame.get_values(block.args)
        while block is not None:
            frame.current_block = block
            frame.set_values(block.args, block_inputs)
            for stmt in block.stmts:
                frame.current_stmt = stmt
                stmt_results = interp_.frame_eval(frame, stmt)
                match stmt_results:
                    case tuple():
                        frame.set_values(stmt._results, stmt_results)
                    case None:
                        continue
                    case interp.Successor(block, block_inputs):
                        pass
                    case interp.ReturnValue():
                        return stmt_results  # terminate the call frame
                    case interp.YieldValue(values):
                        return values  # terminate the region
        return


@dialect.register(key="abstract")
class Abstract(interp.MethodTable):

    FrameType = TypeVar("FrameType", bound=interp.AbstractFrame)
    LatticeType = TypeVar("LatticeType", bound=lattice.BoundedLattice)

    @interp.impl(ir.SSACFG())
    def ssacfg(
        self,
        interp_: interp.AbstractInterpreter[FrameType, LatticeType],
        frame: FrameType,
        node: ir.Region,
    ):
        result = None
        frame.worklist.append(
            interp.Successor(node.blocks[0], *frame.get_values(node.blocks[0].args))
        )
        while (succ := frame.worklist.pop()) is not None:
            visited = frame.visited.setdefault(succ.block, set())
            if succ in visited:
                continue

            block_result = self.run_succ(interp_, frame, succ)
            if len(frame.visited[succ.block]) < 128:
                frame.visited[succ.block].add(succ)
            else:
                continue

            if isinstance(block_result, interp.Successor):
                raise interp.InterpreterError(
                    "unexpected successor, successors should be in worklist"
                )

            result = interp_.join_results(result, block_result)

        if isinstance(result, interp.YieldValue):
            return result.values
        return result

    def run_succ(
        self,
        interp_: interp.AbstractInterpreter[FrameType, LatticeType],
        frame: FrameType,
        succ: interp.Successor,
    ) -> interp.SpecialValue[LatticeType]:
        frame.current_block = succ.block
        frame.set_values(succ.block.args, succ.block_args)
        for stmt in succ.block.stmts:
            frame.current_stmt = stmt
            stmt_results = interp_.frame_eval(frame, stmt)
            if isinstance(stmt_results, tuple):
                frame.set_values(stmt._results, stmt_results)
            elif stmt_results is None:
                continue  # empty result
            else:  # terminate
                return stmt_results
        return None



"""The unpack dialect for Python.

This module contains the dialect for the Python unpack semantics, including:

- The `Unpack` statement class.
- The lowering pass for the unpack statement.
- The concrete implementation of the unpack statement.
- The type inference implementation of the unpack statement.
- A helper function `unpacking` for unpacking Python AST nodes during lowering.
"""

import ast

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.print import Printer

dialect = ir.Dialect("py.unpack")


@statement(dialect=dialect, init=False)
class Unpack(ir.Statement):
    value: ir.SSAValue = info.argument(types.Any)
    names: tuple[str | None, ...] = info.attribute()

    def __init__(self, value: ir.SSAValue, names: tuple[str | None, ...]):
        result_types = [types.Any] * len(names)
        super().__init__(
            args=(value,),
            result_types=result_types,
            args_slice={"value": 0},
            attributes={"names": ir.PyAttr(names)},
        )
        for result, name in zip(self.results, names):
            result.name = name

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.print(self.value)


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Unpack)
    def unpack(self, interp: interp.Interpreter, frame: interp.Frame, stmt: Unpack):
        return tuple(frame.get(stmt.value))


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Unpack)
    def unpack(self, interp, frame: interp.Frame[types.TypeAttribute], stmt: Unpack):
        value = frame.get(stmt.value)
        if isinstance(value, types.Generic) and value.is_subseteq(types.Tuple):
            if value.vararg:
                rest = tuple(value.vararg.typ for _ in stmt.names[len(value.vars) :])
                return tuple(value.vars) + rest
            else:
                return value.vars
        # TODO: support unpacking other types
        return tuple(types.Any for _ in stmt.names)


def unpacking(state: lowering.State, node: ast.expr, value: ir.SSAValue):
    if isinstance(node, ast.Name):
        state.current_frame.defs[node.id] = value
        value.name = node.id
        return
    elif not isinstance(node, ast.Tuple):
        raise lowering.BuildError(f"unsupported unpack node {node}")

    names: list[str | None] = []
    continue_unpack: list[int] = []
    for idx, item in enumerate(node.elts):
        if isinstance(item, ast.Name):
            names.append(item.id)
        else:
            names.append(None)
            continue_unpack.append(idx)
    stmt = state.current_frame.push(Unpack(value, tuple(names)))
    for name, result in zip(names, stmt.results):
        if name is not None:
            state.current_frame.defs[name] = result

    for idx in continue_unpack:
        unpacking(state, node.elts[idx], stmt.results[idx])



from kirin import types, interp
from kirin.analysis import ForwardFrame, TypeInference
from kirin.dialects import func
from kirin.dialects.eltype import ElType

from . import absint
from .stmts import For, IfElse
from ._dialect import dialect


@dialect.register(key="typeinfer")
class TypeInfer(absint.Methods):

    @interp.impl(IfElse)
    def if_else_(
        self,
        interp_: TypeInference,
        frame: ForwardFrame[types.TypeAttribute],
        stmt: IfElse,
    ):
        frame.set(
            stmt.cond, frame.get(stmt.cond).meet(types.Bool)
        )  # set cond backwards
        return super().if_else(interp_, frame, stmt)

    @interp.impl(For)
    def for_loop(
        self,
        interp_: TypeInference,
        frame: ForwardFrame[types.TypeAttribute],
        stmt: For,
    ):
        loop_vars = frame.get_values(stmt.initializers)
        body_block = stmt.body.blocks[0]
        block_args = body_block.args

        eltype = interp_.frame_eval(frame, ElType(stmt.iterable))
        if not isinstance(eltype, tuple):  # error
            return
        item = eltype[0]
        frame.set_values(block_args, (item,) + loop_vars)

        if isinstance(body_block.last_stmt, func.Return):
            frame.worklist.append(interp.Successor(body_block, item, *loop_vars))
            return  # if terminate is Return, there is no result

        loop_vars_ = interp_.frame_call_region(frame, stmt, stmt.body, item, *loop_vars)
        if isinstance(loop_vars_, interp.ReturnValue):
            return loop_vars_
        elif isinstance(loop_vars_, tuple):
            return interp_.join_results(loop_vars, loop_vars_)
        else:  # None, loop has no result
            return



from pytest import mark

from kirin import ir
from kirin.passes import Fold
from kirin.prelude import python_basic
from kirin.dialects import scf, func, lowering

# TODO:
# test_cons
# cond = py.Constant(True)
# then_body = ir.Region(ir.Block())
# else_body = ir.Region(ir.Block())

# ifelse = scf.IfElse(cond.result, ir.Block([scf.Yield(cond.result)]), ir.Block([scf.Yield(cond.result)]))
# ifelse.print()

# ir.Block([ifelse]).print()


@ir.dialect_group(python_basic.union([func, scf, lowering.func]))
def kernel(self):
    def run_pass(method):
        pass

    return run_pass


@mark.xfail(reason="if with early return not supported in scf lowering")
def test_basic_if_else():
    @kernel
    def main(x):
        if x > 0:
            y = x + 1
            z = y + 1
            return z
        else:
            y = x + 2
            z = y + 2

        if x < 0:
            y = y + 3
            z = y + 3
        else:
            y = x + 4
            z = y + 4
        return y, z

    main.print()
    print(main(1))


def test_if_else_defs():

    @kernel
    def main(n: int):
        x = 0

        if x == n:
            x = 1
        else:
            y = 2  # noqa: F841

        return x

    main.print()

    # make sure fold doesn't remove the nested def
    main2 = main.similar(kernel)
    Fold(main2.dialects)(main2)

    assert main(0) == 1 == main2(0)
    assert main(10) == 0 == main2(4)

    main2.print()

    @kernel
    def main_elif(n: int):
        x = 0

        if x == n:
            x = 3
        elif x == n + 1:
            x = 4

        return x

    main_elif.print()

    main_elif2 = main_elif.similar(kernel)
    Fold(main_elif2.dialects)(main_elif2)

    main_elif2.print()

    assert main_elif(0) == 3 == main_elif2(0)
    assert main_elif(-1) == 4 == main_elif2(-1)
    assert main_elif(5) == 0 == main_elif2(7)

    @kernel
    def main_nested_if(n: int):
        x = 0

        if n > 0:
            if n > 1:
                if n == 3:
                    x = 4

        return x

    main_nested_if.print()

    main_nested_if2 = main_nested_if.similar(kernel)
    Fold(main_nested_if2.dialects)(main_nested_if2)

    assert main_nested_if(3) == 4 == main_nested_if2(3)
    assert main_nested_if(10) == 0 == main_nested_if2(8)


@mark.xfail(reason="if with early return not supported in scf lowering")
def test_def_only_else():
    @kernel
    def main(n: int):
        c = 1.0
        if n <= 0:
            return 0.0
        else:
            c = 2.0
        return c

    main.print()
    assert main(1) == 2.0
    assert main(0) == 0.0


@mark.xfail(reason="if with early return not supported in scf lowering")
def test_def_only_else_nested():
    @kernel
    def main(n: int):
        c = 1.0
        if n <= 0:
            return 0.0
        else:
            if n <= 2:
                return 1.0
            else:
                c = 4.0
        return c

    main.print()
    assert main(3) == 4.0
    assert main(2) == 1.0
    assert main(1) == 1.0
    assert main(0) == 0.0



"""Base dialect for Python.

This dialect does not contain statements. It only contains
lowering rules for `ast.Name` and `ast.Expr`.
"""

import ast

from kirin import ir, interp, lowering

dialect = ir.Dialect("py.base")


@dialect.register
class PythonLowering(lowering.FromPythonAST):

    def lower_Name(self, state: lowering.State, node: ast.Name) -> lowering.Result:
        name = node.id
        if isinstance(node.ctx, ast.Load):
            value = state.current_frame.get(name)
            if value is None:
                raise lowering.BuildError(f"{name} is not defined")
            return value
        elif isinstance(node.ctx, ast.Store):
            raise lowering.BuildError("unhandled store operation")
        else:  # Del
            raise lowering.BuildError("unhandled del operation")

    def lower_Expr(self, state: lowering.State, node: ast.Expr) -> lowering.Result:
        return state.parent.visit(state, node.value)


@dialect.register(key="emit.julia")
class PyAttrMethod(interp.MethodTable):

    @interp.impl(ir.PyAttr)
    def py_attr(self, interp, frame: interp.Frame, node: ir.PyAttr):
        return repr(node.data)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar
from dataclasses import dataclass

from kirin import ir
from kirin.interp import Frame, abc

TargetType = TypeVar("TargetType")


@dataclass
class EmitFrame(Frame[TargetType]):
    pass


CodeGenFrameType = TypeVar("CodeGenFrameType", bound=EmitFrame)


@dataclass
class EmitABC(abc.InterpreterABC[CodeGenFrameType, TargetType], ABC):

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for each in getattr(cls, "keys", ()):
            if not each.startswith("emit."):
                raise ValueError(f"Key {each} cannot start with 'emit.'")

    @abstractmethod
    def run(self, node: ir.Method | ir.Statement): ...



from kirin.prelude import basic
from kirin.dialects import ilist
from kirin.serialization.jsonserializer import JSONSerializer
from kirin.serialization.base.serializer import Serializer
from kirin.serialization.base.deserializer import Deserializer


@basic
def foo(x: int, y: float, z: bool):
    c = [[(200.0, 200.0), (210.0, 200.0)]]
    if z:
        c.append([(222.0, 333.0)])
    else:
        return [1, 2, 3, 4]
    return c


@basic
def bar():
    def goo(x: int):
        a = (3, 4)
        return a[0]

    def boo(y):
        return goo(y) + 1

    boo(4)


@basic
def loop_ilist():
    a = 0
    c = ilist.IList([a, a * 2])
    for i in range(3):
        a = i
        c = ilist.IList([a, a * 2])
    return c


@basic
def my_kernel1(x: int):
    return (x, x + 1, 3)


@basic
def my_kernel2(y: int):
    return my_kernel1(y) * 10


@basic
def slicing():
    in1 = ("a", "b", "c", "d", "e", "f", "g", "h")
    in2 = [1, 2, 3, 4, 5]

    x = slice(3, 5)
    a = in2[x]
    b = in1[1:4]
    c = in1[:3]
    d = in1[2:]
    e = in1[:]
    return (a, b, c, d, e)


def round_trip(program):
    serializer = Serializer()
    deserializer = Deserializer()
    encoded = serializer.encode(program)
    decoded = deserializer.decode(encoded)
    assert decoded.code.is_structurally_equal(program.code)
    json_serializer = JSONSerializer()
    json_encoded = json_serializer.encode(encoded)
    json_decoded = json_serializer.decode(json_encoded)
    decoded_2 = deserializer.decode(json_decoded)
    assert decoded_2.code.is_structurally_equal(program.code)


def test_round_trip1():
    round_trip(foo)


def test_round_trip2():
    round_trip(bar)


def test_round_trip3():
    round_trip(loop_ilist)


def test_round_trip4():
    round_trip(my_kernel2)


def test_round_trip5():
    round_trip(slicing)


def test_deterministic():
    serializer = Serializer()
    s1 = serializer.encode(loop_ilist)
    json_serializer = JSONSerializer()
    json_s1 = json_serializer.encode(s1)
    serializer2 = Serializer()
    s2 = serializer2.encode(loop_ilist)
    json_serializer2 = JSONSerializer()
    json_s2 = json_serializer2.encode(s2)
    assert json_s1 == json_s2



"""This module provides a singleton class `Undefined` and `UndefinedType`
that represents an undefined value in the Kirin interpreter.

The `Undefined` class is a singleton that can be used to represent an
undefined value in the interpreter. It is used to indicate that a value
has not been set or is not available. This is used to distinguish between
an undefined value and a Python `None` value.
"""

from typing_extensions import TypeIs


class UndefinedMeta(type):

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        cls._instance = None

    def __call__(cls):
        if cls._instance is None:
            cls._instance = super().__call__()
        return cls._instance


class UndefinedType(metaclass=UndefinedMeta):
    pass


Undefined = UndefinedType()
"""Singleton instance of `UndefinedType` that represents an undefined value."""


def is_undefined(value: object) -> TypeIs[UndefinedType]:
    """Check if the given value is an instance of `UndefinedType`.

    Args:
        value (object): The value to check.

    Returns:
        bool: True if the value is an instance of `UndefinedType`, False otherwise.
    """
    return value is Undefined
